# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Directed edges, as used by LiSE."""
from collections import Mapping, ValuesView

from allegedb.graph import Edge
from allegedb.cache import HistoryError

from .util import getatt
from .query import StatusAlias
from .rule import RuleFollower
from .rule import RuleMapping as BaseRuleMapping


class RuleMapping(BaseRuleMapping):
    """Mapping to get rules followed by a portal."""

    def __init__(self, portal):
        """Store portal, engine, and rulebook."""
        super().__init__(portal.engine, portal.rulebook)
        self.portal = portal


class Portal(Edge, RuleFollower):
    """Connection between two Places that Things may travel along.

    Portals are one-way, but you can make one appear two-way by
    setting the ``symmetrical`` key to ``True``,
    eg. ``character.add_portal(orig, dest, symmetrical=True)``.
    The portal going the other way will appear to have all the
    stats of this one, and attempting to set a stat on it will
    set it here instead.

    """
    __slots__ = ('graph', 'orig', 'dest', 'idx')
    character = getatt('graph')
    engine = getatt('db')
    no_unwrap = True

    @property
    def _cache(self):
        return self.db._edge_val_cache[
            self.character.name][self.orig][self.dest][0]

    def _rule_name_activeness(self):
        rulebook_name = self._get_rulebook_name()
        cache = self.engine._active_rules_cache
        if rulebook_name not in cache:
            return
        cache = cache[rulebook_name]
        for rule in cache:
            for (branch, turn, tick) in self.engine._iter_parent_btt():
                if branch not in cache[rule]:
                    continue
                try:
                    yield (rule, cache[rule][branch][turn][tick])
                    break
                except ValueError:
                    continue
                except HistoryError as ex:
                    if ex.deleted:
                        break
        raise KeyError("{}->{} has no rulebook?".format(
            self.orig, self.dest
        ))

    def _get_rulebook_name(self):
        try:
            return self.engine._portals_rulebooks_cache.retrieve(
                self.character.name, self.orig, self.dest, *self.engine.btt()
            )
        except KeyError:
            return (self.character.name, self.orig, self.dest)

    def _set_rulebook_name(self, rulebook):
        character = self.character
        orig = self.orig
        dest = self.dest
        cache = self.engine._portals_rulebooks_cache
        try:
            if rulebook == cache.retrieve(character, orig, dest, *self.engine.btt()):
                return
        except KeyError:
            pass
        branch, turn, tick = self.engine.nbtt()
        cache.store(character, orig, dest, branch, turn, tick, rulebook)
        self.engine.query.set_portal_rulebook(character, orig, dest, branch, turn, tick, rulebook)

    def _get_rule_mapping(self):
        return RuleMapping(self)

    def __getitem__(self, key):
        """Get the present value of the key.

        If I am a mirror of another Portal, return the value from that
        Portal instead.

        """
        if key == 'origin':
            return self.orig
        elif key == 'destination':
            return self.dest
        elif key == 'character':
            return self.character.name
        elif key == 'is_mirror':
            try:
                return super().__getitem__(key)
            except KeyError:
                return False
        elif 'is_mirror' in self and self['is_mirror']:
            return self.character.preportal[
                self.orig
            ][
                self.dest
            ][
                key
            ]
        else:
            return super().__getitem__(key)

    def __setitem__(self, key, value):
        """Set ``key``=``value`` at the present game-time.

        If I am a mirror of another Portal, set ``key``==``value`` on
        that Portal instead.

        """
        if key in ('origin', 'destination', 'character'):
            raise KeyError("Can't change " + key)
        elif 'is_mirror' in self and self['is_mirror']:
            self.reciprocal[key] = value
            return
        elif key == 'symmetrical' and value:
            if (
                    self.dest not in self.character.portal or
                    self.orig not in
                    self.character.portal[self.dest]
            ):
                self.character.add_portal(self.dest, self.orig)
            self.character.portal[
                self.dest
            ][
                self.orig
            ][
                "is_mirror"
            ] = True
            self.send(self, key='symmetrical', val=False)
            return
        elif key == 'symmetrical' and not value:
            try:
                self.character.portal[
                    self.dest
                ][
                    self.orig
                ][
                    "is_mirror"
                ] = False
            except KeyError:
                pass
            self.send(self, key='symmetrical', val=False)
            return
        super().__setitem__(key, value)

    def __repr__(self):
        """Describe character, origin, and destination"""
        return "{}.character[{}].portal[{}][{}]".format(
            repr(self.engine),
            repr(self['character']),
            repr(self['origin']),
            repr(self['destination'])
        )

    def __bool__(self):
        """It means something that I exist, even if I have no data."""
        return self.orig in self.character.portal and \
               self.dest in self.character.portal[self.orig]

    @property
    def origin(self):
        """Return the Place object that is where I begin"""
        return self.character.place[self.orig]

    @property
    def destination(self):
        """Return the Place object at which I end"""
        return self.character.place[self.dest]

    @property
    def reciprocal(self):
        """If there's another Portal connecting the same origin and
        destination that I do, but going the opposite way, return
        it. Else raise KeyError.

        """
        try:
            return self.character.portal[self.dest][self.orig]
        except KeyError:
            raise KeyError("This portal has no reciprocal")

    def historical(self, stat):
        return StatusAlias(
            entity=self,
            stat=stat
        )

    def update(self, d):
        """Works like regular update, but only actually updates when the new
        value and the old value differ. This is necessary to prevent
        certain infinite loops.

        """
        for (k, v) in d.items():
            if k not in self or self[k] != v:
                self[k] = v

    def delete(self):
        """Remove myself from my :class:`Character`.

        For symmetry with :class:`Thing` and :class`Place`.

        """
        branch, turn, tick = self.engine.nbtt()
        self.engine._edges_cache.store(
            self.character.name,
            self.origin.name,
            self.destination.name,
            0,
            branch,
            turn,
            tick,
            None
        )
        self.engine.query.exist_edge(
            self.character.name,
            self.origin.name,
            self.destination.name,
            branch, turn, tick, False
        )
        try:
            del self.engine._portal_objs[
                (self.graph.name, self.orig, self.dest)
            ]
        except KeyError:
            pass
        self.character.portal[self.origin.name].send(
            self.character.portal[self.origin.name],
            key='dest', val=None
        )

    def unwrap(self):
        return {
            k: v.unwrap() if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap')
            else v for (k, v) in self.items()
        }


# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The type of node that is a location.

Though both things and places are nodes, things are obliged to be
located in another node. Places are not.

"""


from .node import Node


class Place(Node):
    """The kind of node where a thing might ultimately be located."""
    __slots__ = ('graph', 'db', 'node')

    extrakeys = {
        'name',
        'character'
    }

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return {'name': self.name, 'character': self.character.name}[key]

    def __repr__(self):
        return "{}.character[{}].place[{}]".format(
            repr(self.engine),
            repr(self['character']),
            repr(self['name'])
        )

    def delete(self):
        """Remove myself from the world model immediately."""
        super().delete()
        self.character.place.send(self.character.place, key=self.name, val=None)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This is an interpreter for Benjamin Lortimer's Timecraft crafting ruleset.

It could be useful if you want a similar crafting system in your game; this way,
you can describe it in the same tabular format.

"""
from LiSE.rule import Rule


def install(engine):
    @engine.trigger
    def verbing(engine, character, verb):
        return character.stat.get('verb', None) == verb

    @engine.action
    def craft_thing_dict(engine, character, thing_dict):
        d = thing_dict
        body = character.avatar.physical
        num_of_kind = 0
        for node in character.node:
            if node.name == d['kind'] + str(num_of_kind):
                num_of_kind += 1
        name = d['kind'] + str(num_of_kind)
        body.new_thing(name, **d)

    @engine.trigger
    @engine.prereq
    def have_number_of_kind(engine, character, number, kind):
        body = character.avatar.physical
        return len(
            thing for thing in body.contents()
            if thing.get('kind', None) == kind
        ) >= number

    @engine.action
    def destroy_number_of_kind(engine, character, number, kind):
        body = character.avatar.physical
        n = 0
        for thing in body.contents():
            if n > number:
                break
            if thing['kind'] != kind:
                continue
            thing.delete()
            n += 1

    @engine.prereq
    def have_skill(engine, character, skill, level=None):
        if level:
            return character.skills[skill] >= level
        else:
            return skill in character.skills

    @engine.prereq
    def have_energy(engine, character, energy):
        return character.stat['energy'] >= energy

    @engine.action
    def deduct_energy(engine, character, energy):
        character.stat['energy'] -= energy

    def make_timecraft_rule(
            engine, name,
            energy=10, thing_dicts_to_add=[], material_reqs=[], skill_reqs=[]
    ):
        material_prereqs = []
        material_destroyers = []
        for req in material_reqs:
            if isinstance(req, tuple):
                if not (
                        len(req) == 2 and
                        isinstance(req[0], int) and
                        isinstance(req[1], str)
                ):
                    raise ValueError("Illegal material req: {}".format(req))
                material_prereqs.append(engine.prereq.partial(
                    'have_number_of_kind', number=req[0], kind=req[1]
                ))
                material_destroyers.append(engine.action.partial(
                    'destroy_number_of_kind', number=req[0], kind=req[1]
                ))
            elif isinstance(req, str):
                material_prereqs.append(engine.prereq.partial(
                    'have_number_of_kind', number=1, kind=req
                ))
                material_destroyers.append(engine.action.partial(
                    'destroy_number_of_kind', number=1, kind=req
                ))
            else:
                raise ValueError("Illegal material req: {}".format(req))
        skill_prereqs = []
        for req in skill_reqs:
            if isinstance(req, tuple):
                if not (
                        len(req) == 2 and
                        isinstance(req[0], int) and
                        isinstance(req[1], str)
                ):
                    raise ValueError("Illegal skill req: {}".format(req))
                skill_prereqs.append(engine.prereq.partial(
                    'have_skill', level=req[0], skill=req[1]
                ))
            elif isinstance(req, str):
                skill_prereqs.append(engine.prereq.partial(
                    'have_skill', skill=req
                ))
            else:
                raise ValueError("Illegal skill req: {}".format(req))
        return Rule(
            engine,
            name,
            triggers=[
                engine.trigger.partial(
                    'verbing', action=name
                )
            ],
            prereqs=material_prereqs + skill_prereqs,
            actions=[
                engine.action.partial(
                    'craft_thing_dict', thing_dict=thingd
                ) for thingd in thing_dicts_to_add
            ] + material_destroyers
        )


if __name__ == '__main__':
    import LiSE
    engine = LiSE.Engine('LiSEworld.db', 'LiSEcode.db')
    install(engine)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A simulation of students and teachers, both living on campus,
attending classes and teaching or learning, as appropriate.

Learning is modeled in a way similar to the game Kudos 2: each
student has 100 "brain cells," and the teacher sends an experience
point to each, once per lesson. If the brain cell is awake and
alert, the student receives the experience point and puts it toward
leveling-up. Otherwise it's wasted.

Some students are slow and some are drunkards. These students will
randomly show up late (brain cells become useless in proportion to
how late, recovering immediately afterward) or drunk (some brain
cells made useless per drink, recovering as time passes).

This script will initialize LiSEworld.db and code libraries to run the
simulation described. To view it, run ELiDE from the same directory
as you ran this script from.

"""


def install(eng):
    phys = eng.new_character('physical')
    phys.stat['hour'] = 0

    @phys.rule(always=True)  # runs every tick regardless of the situation
    def time_passes(character):
        character.stat['hour'] = (character.stat['hour'] + 1) % 24


    # There's a character with all of the students in it, to make it easy to apply rules to all students.
    student_body = eng.new_character('student_body')


    classroom = phys.new_place('classroom')


    @student_body.avatar.rule
    def go_to_class(node):
        # There's just one really long class every day.
        node.travel_to(node.character.place['classroom'])


    @go_to_class.trigger
    def absent(node):
        return node.location != node.character.place['classroom']


    @go_to_class.prereq
    def class_in_session(node):
        return 8 <= node.engine.character['physical'].stat['hour'] < 15


    @go_to_class.prereq
    def be_timely(node):
        # Even lazy students have a 50% chance of going to class every hour.
        #
        # We need to access the student character like this because the
        # ``character`` passed into the function is ``student_body``, where the
        # rule was assigned.
        #
        # Or we could have put the 'lazy' stat onto the node instead of the
        # character... or kept the student character in a stat of the node...
        # or assigned this rule to the student directly.
        for user in node.users.values():
            if user.name not in ('physical', 'student_body'):
                return not user.stat['lazy'] or node.engine.coinflip()


    @student_body.avatar.rule
    def leave_class(node):
        for user in node.users.values():
            if user.name != 'student_body':
                node.travel_to(user.stat['room'])
                return


    @leave_class.trigger
    def in_classroom_after_class(node):
        phys = node.character
        return node.location == phys.place['classroom'] \
               and phys.stat['hour'] >= 15


    # Let's make some rules and not assign them to anything yet.
    @eng.rule
    def drink(character):
        braincells = list(character.node.values())
        character.engine.shuffle(braincells)
        for i in range(0, character.engine.randrange(1, 20)):
            braincells.pop()['drunk'] += 12


    @drink.trigger
    def party_time(character):
        phys = character.engine.character['physical']
        return 23 >= phys.stat['hour'] > 15


    @drink.prereq
    def is_drunkard(character):
        return character.stat['drunkard']


    @eng.rule
    def sloth(character):
        braincells = list(character.node.values())
        character.engine.shuffle(braincells)
        for i in range(0, character.engine.randrange(1, 20)):
            braincells.pop()['slow'] += 1


    @sloth.trigger
    def out_of_class(character):
        # You don't want to use the global variable for the classroom
        # because it won't be around (or at least, won't work) after
        # the engine restarts.
        avatar = character.avatar['physical'].only
        classroom = avatar.character.place['classroom']
        return avatar.location != classroom

    sloth.prereq(class_in_session)


    @eng.rule
    def learn(node):
        for user in node.users.values():
            if 'xp' in user.stat:
                user.stat['xp'] += 1


    @learn.trigger
    def in_class(node):
        classroom = node.engine.character['physical'].place['classroom']
        if hasattr(node, 'location'):
            assert node == node.character.avatar['physical'].only
            return node.location == classroom
        else:
            return node.character.avatar['physical'].only.location == classroom

    learn.prereq(class_in_session)


    @learn.prereq
    def pay_attention(node):
        return node['drunk'] == node['slow'] == 0


    @eng.rule
    def sober_up(node):
        node['drunk'] -= 1


    @sober_up.trigger
    def somewhat_drunk(node):
        return node['drunk'] > 0


    @eng.rule
    def catch_up(node):
        node['slow'] -= 1


    @catch_up.trigger
    def somewhat_late(node):
        return node['slow'] > 0

    catch_up.prereq(in_class)
    catch_up.prereq(class_in_session)

    # 3 dorms of 12 students each.
    # Each dorm has 6 rooms.
    # Modeling the teachers would be a logical way to extend this.
    student_body.stat['characters'] = []
    for n in range(0, 3):
        dorm = eng.new_character('dorm{}'.format(n))
        common = phys.new_place('common{}'.format(n))  # A common room for students to meet in
        dorm.add_avatar(common)
        common.two_way(classroom)
        # All rooms in a dorm are connected via its common room
        for i in range(0, 6):
            room = phys.new_place('dorm{}room{}'.format(n, i))
            dorm.add_avatar(room)
            room.two_way(common)
            student0 = eng.new_character('dorm{}room{}student0'.format(n, i))
            body0 = room.new_thing('dorm{}room{}student0'.format(n, i))
            student0.add_avatar(body0)
            assert (student0 in body0.users.values())
            student_body.add_avatar(body0)
            assert (student_body in body0.users.values())
            assert (student0 in body0.users.values())
            student1 = eng.new_character('dorm{}room{}student1'.format(n, i))
            body1 = room.new_thing('dorm{}room{}student1'.format(n, i))
            student1.add_avatar(body1)
            student_body.add_avatar(body1)
            student0.stat['room'] = student1.stat['room'] = room
            assert student0.stat['room'] == student1.stat['room'] == room
            student0.stat['roommate'] = student1
            student1.stat['roommate'] = student0
            for student in (student0, student1):
                if student not in student_body.stat['characters']:
                    student_body.stat['characters'].append(student)
                # Students' nodes are their brain cells.
                # They are useless if drunk or slow, but recover from both conditions a bit every hour.
                for k in range(0, 100):
                    cell = student.new_node('cell{}'.format(k), drunk=0, slow=0)
                    #  ``new_node`` is just an alias for ``new_place``;
                    #  perhaps more logical when the places don't really
                    #  represent potential locations
                    student.stat['xp'] = 0
                    student.stat['drunkard'] = eng.coinflip()
                    student.stat['lazy'] = eng.coinflip()
                # Apply these previously written rules to each student
                for rule in (drink, sloth):
                    student.rule(rule)
                # Apply these previously written rules to each brain cell
                for rule in (learn, sober_up, catch_up):
                    student.place.rule(rule)


if __name__ == "__main__":
    from LiSE.engine import Engine
    with Engine(":memory:") as eng:
        install(eng)
        for i in range(72):
            eng.next_turn()
            print(i)



# Parable of the Polygons is public domain.
# This implementation is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Implementation of Parable of the Polygons http://ncase.me/polygons/"""


def install(eng):
    @eng.function
    def cmp_neighbor_shapes(poly, cmp, stat):
        """Compare the proportion of neighboring polys with the same shape as this one

        Count the neighboring polys that are the same shape as this one, and return how that compares with
        some stat on the poly's user.

        """
        from operator import attrgetter
        home = poly.location
        similar = 0
        n = 0
        for n, neighbor_home in enumerate(map(attrgetter('destination'), home.portal.values()), 1):
            # assume only 1 poly per home for now; this is faithful to the original
            try:
                neighbor = next(iter(neighbor_home.contents()))
            except StopIteration:
                continue
            if neighbor.user is poly.user:
                similar += 1
        return cmp(poly.character.stat[stat], similar / n)

    @eng.rule
    def relocate(poly):
        """Move to a random unoccupied place"""
        unoccupied = [place for place in poly.character.place.values() if not place.content]
        poly.location = poly.engine.choice(unoccupied)

    @relocate.trigger
    def similar_neighbors(poly):
        """Trigger when my neighborhood fails to be enough like me"""
        from operator import ge
        return poly.engine.function.cmp_neighbor_shapes(poly, ge, 'min_sameness')

    @relocate.trigger
    def dissimilar_neighbors(poly):
        """Trigger when my neighborhood gets too much like me"""
        from operator import lt
        return poly.engine.function.cmp_neighbor_shapes(poly, lt, 'max_sameness')


    eng.rulebook['parable'] = [relocate]


    physical = eng.new_character(
        'physical', min_sameness=.1, max_sameness=.9,
        _config={
            'min_sameness': {'control': 'slider', 'min': 0.0, 'max': 1.0},
            'max_sameness': {'control': 'slider', 'min': 0.0, 'max': 1.0}
        }
    )
    square = eng.new_character('square')
    triangle = eng.new_character('triangle')
    square.avatar.rulebook = triangle.avatar.rulebook = 'parable'


    # make an 8-way-connected grid
    physical.grid_2d_8graph(20, 20)
    empty = list(physical.place.values())
    eng.shuffle(empty)
    # distribute 30 of each shape randomly among the empty places
    for i in range(1, 31):
        square.add_avatar(empty.pop().new_thing('square%i' % i, _image_paths=['atlas://polygons/meh_square']))
    for i in range(1, 31):
        triangle.add_avatar(empty.pop().new_thing('triangle%i' % i, _image_paths=['atlas://polygons/meh_triangle']))


if __name__ == '__main__':
    import os
    from LiSE import Engine
    for stale in ('LiSEworld.db', 'trigger.py', 'prereq.py', 'action.py', 'function.py', 'method.py'):
        if os.path.exists(stale):
            os.remove(stale)
    with Engine('LiSEworld.db') as eng:
        with eng.batch():
            install(eng)
    import sys
    if '--profile' in sys.argv:
        import cProfile

        def test():
            with Engine('LiSEworld.db') as eng:
                for n in range(10):
                    eng.next_turn()

        cProfile.run('test()', 'polygons.prof')


# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector,  public@zacharyspector.com
"""The "engine" of LiSE is an object relational mapper with special
stores for game data and entities, as well as properties for manipulating the
flow of time.

"""
from functools import partial
from collections import defaultdict
from operator import attrgetter
from types import FunctionType, MethodType

import umsgpack
from blinker import Signal
from allegedb import ORM as gORM
from .util import reify, sort_set

from . import exc


class NoPlanningAttrGetter:
    __slots__ = ('_real',)

    def __init__(self, attr, *attrs):
        self._real = attrgetter(attr, *attrs)

    def __call__(self, obj):
        if obj._planning:
            raise exc.PlanError("Don't use randomization in a plan")
        return self._real(obj)


def getnoplan(attribute_name):
    return property(NoPlanningAttrGetter(attribute_name))


class InnerStopIteration(StopIteration):
    pass


class NextTurn(Signal):
    """Make time move forward in the simulation.

    Calls ``advance`` repeatedly, returning a list of the rules' return values.

    I am also a ``Signal``, so you can register functions to be
    called when the simulation runs. Pass them to my ``connect``
    method.

    """
    def __init__(self, engine):
        super().__init__()
        self.engine = engine

    def __call__(self):
        engine = self.engine
        start_branch, start_turn, start_tick = engine.btt()
        latest_turn = engine._turns_completed[start_branch]
        if start_turn < latest_turn:
            engine.turn += 1
            self.send(
                engine,
                branch=engine.branch,
                turn=engine.turn,
                tick=engine.tick
            )
            return [], engine.get_delta(
                branch=start_branch,
                turn_from=start_turn,
                turn_to=engine.turn,
                tick_from=start_tick,
                tick_to=engine.tick
            )
        elif start_turn > latest_turn + 1:
            raise exc.RulesEngineError("Can't run the rules engine on any turn but the latest")
        if start_turn == latest_turn:
            # As a side effect, the following assignment sets the tick to
            # the latest in the new turn, which will be 0 if that turn has not
            # yet been simulated.
            engine.turn += 1
            if engine.tick == 0:
                engine.universal['rando_state'] = engine._rando.getstate()
            else:
                engine._rando.setstate(engine.universal['rando_state'])
        with engine.advancing():
            for res in iter(engine.advance, final_rule):
                if res:
                    engine.universal['last_result'] = res
                    engine.universal['last_result_idx'] = 0
                    engine.universal['rando_state'] = engine._rando.getstate()
                    branch, turn, tick = engine.btt()
                    self.send(
                        engine,
                        branch=branch,
                        turn=turn,
                        tick=tick
                    )
                    return res, engine.get_delta(
                        branch=start_branch,
                        turn_from=start_turn,
                        turn_to=turn,
                        tick_from=start_tick,
                        tick_to=tick
                    )
        engine._turns_completed[start_branch] = engine.turn
        engine.query.complete_turn(start_branch, engine.turn)
        self.send(
            self.engine,
            branch=engine.branch,
            turn=engine.turn,
            tick=engine.tick
        )
        return [], engine.get_delta(
            branch=engine.branch,
            turn_from=start_turn,
            turn_to=engine.turn,
            tick_from=start_tick,
            tick_to=engine.tick
        )


class DummyEntity(dict):
    """Something to use in place of a node or edge"""
    __slots__ = ['engine']

    def __init__(self, engine):
        self.engine = engine


class FinalRule:
    """A singleton sentinel for the rule iterator"""
    __slots__ = []

    def __hash__(self):
        # completely random integer
        return 6448962173793096248


final_rule = FinalRule()


MSGPACK_TUPLE = 0x00
MSGPACK_FROZENSET = 0x01
MSGPACK_SET = 0x02
MSGPACK_EXCEPTION = 0x03
MSGPACK_CHARACTER = 0x7f
MSGPACK_PLACE = 0x7e
MSGPACK_THING = 0x7d
MSGPACK_PORTAL = 0x7c
MSGPACK_FINAL_RULE = 0x7b
MSGPACK_FUNCTION = 0x7a
MSGPACK_METHOD = 0x79
MSGPACK_TRIGGER = 0x78
MSGPACK_PREREQ = 0x77
MSGPACK_ACTION = 0x76


class AbstractEngine(object):
    """Parent class to the real Engine as well as EngineProxy.

    Implements serialization methods and the __getattr__ for stored methods.

    By default, the deserializers will refuse to create LiSE entities. If
    you want them to, use my ``loading`` property to open a ``with`` block,
    in which deserialized entities will be created as needed.

    """
    from contextlib import contextmanager

    @contextmanager
    def loading(self):
        """Context manager for when you need to instantiate entities upon unpacking"""
        if getattr(self, '_initialized', False):
            raise ValueError("Already loading")
        self._initialized = False
        yield
        self._initialized = True

    def __getattr__(self, item):
        meth = super().__getattribute__('method').__getattr__(item)
        return MethodType(meth, self)

    def _pack_character(self, char):
        return umsgpack.Ext(MSGPACK_CHARACTER, umsgpack.packb(char.name, ext_handlers=self._pack_handlers))

    def _pack_place(self, place):
        return umsgpack.Ext(MSGPACK_PLACE, umsgpack.packb(
            (place.character.name, place.name), ext_handlers=self._pack_handlers
        ))

    def _pack_thing(self, thing):
        return umsgpack.Ext(MSGPACK_THING, umsgpack.packb(
            (thing.character.name, thing.name), ext_handlers=self._pack_handlers
        ))

    def _pack_portal(self, port):
        return umsgpack.Ext(MSGPACK_PORTAL, umsgpack.packb(
            (port.character.name, port.orig, port.dest), ext_handlers=self._pack_handlers
        ))

    def _pack_tuple(self, tup):
        return umsgpack.Ext(MSGPACK_TUPLE, umsgpack.packb(list(tup), ext_handlers=self._pack_handlers))

    def _pack_frozenset(self, frozs):
        return umsgpack.Ext(MSGPACK_FROZENSET, umsgpack.packb(list(frozs), ext_handlers=self._pack_handlers))

    def _pack_set(self, s):
        return umsgpack.Ext(MSGPACK_SET, umsgpack.packb(list(s), ext_handlers = self._pack_handlers))

    def _pack_exception(self, exc):
        return umsgpack.Ext(MSGPACK_EXCEPTION, umsgpack.packb(
            [exc.__class__.__name__] + list(exc.args), ext_handlers=self._pack_handlers
        ))

    def _pack_func(self, func):
        return umsgpack.Ext({
            'method': MSGPACK_METHOD,
            'function': MSGPACK_FUNCTION,
            'trigger': MSGPACK_TRIGGER,
            'prereq': MSGPACK_PREREQ,
            'action': MSGPACK_ACTION
        }[func.__module__], umsgpack.packb(func.__name__))

    def _pack_meth(self, func):
        return umsgpack.Ext(MSGPACK_METHOD, umsgpack.packb(func.__name__))

    def _unpack_char(self, ext):
        charn = umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers)
        try:
            return self.character[charn]
        except KeyError:
            if getattr(self, '_initialized', True):
                raise
            return self.char_cls(self, charn)

    def _unpack_place(self, ext):
        charn, placen = umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers)
        try:
            char = self.character[charn]
        except KeyError:
            if getattr(self, '_initialized', True):
                raise
            return self.place_cls(self.char_cls(self, charn), placen)
        try:
            return char.place[placen]
        except KeyError:
            if getattr(self, '_initialized', True):
                raise
            return self.place_cls(char, placen)

    def _unpack_thing(self, ext):
        charn, thingn = umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers)
        try:
            char = self.character[charn]
        except KeyError:
            if getattr(self, '_initialized', True):
                raise
            return self.thing_cls(self.char_cls(self, charn), thingn)
        try:
            return char.thing[thingn]
        except KeyError:
            if getattr(self, '_initialized', True):
                raise
            return self.thing_cls(char, thingn)

    def _unpack_portal(self, ext):
        charn, orign, destn = umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers)
        try:
            char = self.character[charn]
        except KeyError:
            if getattr(self, '_initialized', True):
                raise
            char = self.char_cls(self, charn)
        try:
            return char.portal[orign][destn]
        except KeyError:
            if getattr(self, '_initialized', True):
                raise
            return self.portal_cls(char, orign, destn)

    def _unpack_trigger(self, ext):
        return getattr(self.trigger, umsgpack.unpackb(ext.data))

    def _unpack_prereq(self, ext):
        return getattr(self.prereq, umsgpack.unpackb(ext.data))

    def _unpack_action(self, ext):
        return getattr(self.action, umsgpack.unpackb(ext.data))

    def _unpack_function(self, ext):
        return getattr(self.function, umsgpack.unpackb(ext.data))

    def _unpack_method(self, ext):
        return getattr(self.method, umsgpack.unpackb(ext.data))

    def _unpack_tuple(self, ext):
        return tuple(umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers))

    def _unpack_frozenset(self, ext):
        return frozenset(umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers))

    def _unpack_set(self, ext):
        return set(umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers))

    def _unpack_exception(self, ext):
        excs = {
            # builtin exceptions
            'AssertionError': AssertionError,
            'AttributeError': AttributeError,
            'EOFError': EOFError,
            'FloatingPointError': FloatingPointError,
            'GeneratorExit': GeneratorExit,
            'ImportError': ImportError,
            'IndexError': IndexError,
            'KeyError': KeyError,
            'KeyboardInterrupt': KeyboardInterrupt,
            'MemoryError': MemoryError,
            'NameError': NameError,
            'NotImplementedError': NotImplementedError,
            'OSError': OSError,
            'OverflowError': OverflowError,
            'RecursionError': RecursionError,
            'ReferenceError': ReferenceError,
            'RuntimeError': RuntimeError,
            'StopIteration': StopIteration,
            'IndentationError': IndentationError,
            'TabError': TabError,
            'SystemError': SystemError,
            'SystemExit': SystemExit,
            'TypeError': TypeError,
            'UnboundLocalError': UnboundLocalError,
            'UnicodeError': UnicodeError,
            'UnicodeEncodeError': UnicodeEncodeError,
            'UnicodeDecodeError': UnicodeDecodeError,
            'UnicodeTranslateError': UnicodeTranslateError,
            'ValueError': ValueError,
            'ZeroDivisionError': ZeroDivisionError,
            # LiSE exceptions
            'NonUniqueError': exc.NonUniqueError,
            'AmbiguousAvatarError': exc.AmbiguousAvatarError,
            'AmbiguousUserError': exc.AmbiguousUserError,
            'RulesEngineError': exc.RulesEngineError,
            'RuleError': exc.RuleError,
            'RedundantRuleError': exc.RedundantRuleError,
            'UserFunctionError': exc.UserFunctionError,
            'WorldIntegrityError': exc.WorldIntegrityError,
            'CacheError': exc.CacheError,
            'TravelException': exc.TravelException
        }
        data = umsgpack.unpackb(ext.data, ext_handlers=self._unpack_handlers)
        if data[0] not in excs:
            return Exception(*data)
        return excs[data[0]](*data[1:])


    @reify
    def _unpack_handlers(self):
        return {
            MSGPACK_CHARACTER: self._unpack_char,
            MSGPACK_PLACE: self._unpack_place,
            MSGPACK_THING: self._unpack_thing,
            MSGPACK_PORTAL: self._unpack_portal,
            MSGPACK_FINAL_RULE: lambda obj: final_rule,
            MSGPACK_TUPLE: self._unpack_tuple,
            MSGPACK_FROZENSET: self._unpack_frozenset,
            MSGPACK_SET: self._unpack_set,
            MSGPACK_TRIGGER: self._unpack_trigger,
            MSGPACK_PREREQ: self._unpack_prereq,
            MSGPACK_ACTION: self._unpack_action,
            MSGPACK_FUNCTION: self._unpack_function,
            MSGPACK_METHOD: self._unpack_method,
            MSGPACK_EXCEPTION: self._unpack_exception
        }

    @reify
    def _pack_handlers(self):
        return {
            self.char_cls: self._pack_character,
            self.place_cls: self._pack_place,
            self.thing_cls: self._pack_thing,
            self.portal_cls: self._pack_portal,
            tuple: self._pack_tuple,
            frozenset: self._pack_frozenset,
            set: self._pack_set,
            FinalRule: lambda obj: umsgpack.Ext(MSGPACK_FINAL_RULE, b""),
            FunctionType: self._pack_func,
            MethodType: self._pack_meth,
            Exception: self._pack_exception
        }

    def pack(self, obj):
        return umsgpack.packb(obj, ext_handlers=self._pack_handlers)

    def unpack(self, bs):
        return umsgpack.unpackb(bs, ext_handlers=self._unpack_handlers)

    def coinflip(self):
        """Return True or False with equal probability."""
        return self.choice((True, False))

    def roll_die(self, d):
        """Roll a die with ``d`` faces. Return the result."""
        return self.randint(1, d)

    def dice(self, n, d):
        """Roll ``n`` dice with ``d`` faces, and yield the results.

        This is an iterator. You'll get the result of each die in
        successon.

        """
        for i in range(0, n):
            yield self.roll_die(d)

    def dice_check(self, n, d, target, comparator='<='):
        """Roll ``n`` dice with ``d`` sides, sum them, and return whether they
        are <= ``target``.

        If ``comparator`` is provided, use it instead of <=. You may
        use a string like '<' or '>='.

        """
        from operator import gt, lt, ge, le, eq, ne

        comps = {
            '>': gt,
            '<': lt,
            '>=': ge,
            '<=': le,
            '=': eq,
            '==': eq,
            '!=': ne
        }
        try:
            comparator = comps.get(comparator, comparator)
        except TypeError:
            pass
        return comparator(sum(self.dice(n, d)), target)

    def percent_chance(self, pct):
        """Given a ``pct``% chance of something happening right now, decide at
        random whether it actually happens, and return ``True`` or
        ``False`` as appropriate.

        Values not between 0 and 100 are treated as though they
        were 0 or 100, whichever is nearer.

        """
        if pct <= 0:
            return False
        if pct >= 100:
            return True
        return pct / 100 < self.random()

    betavariate = getnoplan('_rando.betavariate')
    choice = getnoplan('_rando.choice')
    expovariate = getnoplan('_rando.expovariate')
    gammavariate = getnoplan('_rando.gammavariate')
    gauss = getnoplan('_rando.gauss')
    getrandbits = getnoplan('_rando.getrandbits')
    lognormvariate = getnoplan('_rando.lognormvariate')
    normalvariate = getnoplan('_rando.normalvariate')
    paretovariate = getnoplan('_rando.paretovariate')
    randint = getnoplan('_rando.randint')
    random = getnoplan('_rando.random')
    randrange = getnoplan('_rando.randrange')
    sample = getnoplan('_rando.sample')
    shuffle = getnoplan('_rando.shuffle')
    triangular = getnoplan('_rando.triangular')
    uniform = getnoplan('_rando.uniform')
    vonmisesvariate = getnoplan('_rando.vonmisesvariate')
    weibullvariate = getnoplan('_rando.weibullvariate')


class Engine(AbstractEngine, gORM):
    """LiSE, the Life Simulator Engine.

    Each instance of LiSE maintains a connection to a database
    representing the state of a simulated world. Simulation rules
    within this world are described by lists of Python functions, some
    of which make changes to the world.

    The top-level data structure within LiSE is the character. Most
    data within the world model is kept in some character or other;
    these will quite frequently represent people, but can be readily
    adapted to represent any kind of data that can be comfortably
    described as a graph or a JSON object. Every change to a character
    will be written to the database.

    LiSE tracks history as a series of turns. In each turn, each
    simulation rule is evaluated once for each of the simulated
    entities it's been applied to. World changes in a given turn are
    remembered together, such that the whole world state can be
    rewound: simply set the properties ``branch`` and ``turn`` back to
    what they were just before the change you want to undo.

    Properties:

    - ``branch``: The fork of the timestream that we're on.
    - ``turn``: Units of time that have passed since the sim started.
    - ``time``: ``(branch, turn)``
    - ``tick``: A counter of how many changes have occurred this turn
    - ``character``: A mapping of :class:`Character` objects by name.
    - ``rule``: A mapping of all rules that have been made.
    - ``rulebook``: A mapping of lists of rules. They are followed in
      their order.  A whole rulebook full of rules may be assigned to
      an entity at once.
    - ``trigger``: A mapping of functions that might trigger a rule.
    - ``prereq``: A mapping of functions a rule might require to return
      ``True`` for it to run.
    - ``action``: A mapping of functions that might manipulate the world
      state as a result of a rule running.
    - ``function``: A mapping of generic functions.
    - ``string``: A mapping of strings, probably shown to the player
      at some point.
    - ``eternal``: Mapping of arbitrary serializable objects. It isn't
      sensitive to sim-time. A good place to keep game settings.
    - ``universal``: Another mapping of arbitrary serializable
      objects, but this one *is* sensitive to sim-time. Each turn, the
      state of the randomizer is saved here under the key
      ``'rando_state'``.

    """
    from .character import Character
    from .thing import Thing
    from .place import Place
    from .portal import Portal
    from .query import QueryEngine
    char_cls = Character
    thing_cls = Thing
    place_cls = node_cls = Place
    portal_cls = edge_cls = Portal
    query_engine_cls = QueryEngine
    illegal_graph_names = ['global', 'eternal', 'universal', 'rulebooks', 'rules']
    illegal_node_names = ['nodes', 'node_val', 'edges', 'edge_val', 'things']

    def _make_node(self, graph, node):
        if self._is_thing(graph.name, node):
            return self.thing_cls(graph, node)
        else:
            return self.place_cls(graph, node)

    def _make_edge(self, graph, orig, dest, idx=0):
        return self.portal_cls(graph, orig, dest)

    def get_delta(self, branch, turn_from, tick_from, turn_to, tick_to):
        """Get a dictionary describing changes to the world.

        Most keys will be character names, and their values will be dictionaries of
        the character's stats' new values, with ``None`` for deleted keys. Characters'
        dictionaries have special keys 'nodes' and 'edges' which contain booleans indicating
        whether the node or edge exists at the moment, and 'node_val' and 'edge_val' for
        the stats of those entities. For edges (also called portals) these dictionaries
        are two layers deep, keyed first by the origin, then by the destination.

        Characters also have special keys for the various rulebooks they have:

        * 'character_rulebook'
        * 'avatar_rulebook'
        * 'character_thing_rulebook'
        * 'character_place_rulebook'
        * 'character_portal_rulebook'

        And each node and edge may have a 'rulebook' stat of its own. If a node is a thing,
        it gets a 'location'; when the 'location' is deleted,
        that means it's back to being a place.

        Keys at the top level that are not character names:

        * 'rulebooks', a dictionary keyed by the name of each changed rulebook, the value
        being a list of rule names
        * 'rules', a dictionary keyed by the name of each changed rule, containing any
        of the lists 'triggers', 'prereqs', and 'actions'

        """
        from allegedb.window import update_window, update_backward_window
        if turn_from == turn_to:
            return self.get_turn_delta(branch, turn_to, tick_to, start_tick=tick_from)
        delta = super().get_delta(branch, turn_from, tick_from, turn_to, tick_to)
        if turn_from < turn_to:
            updater = partial(update_window, turn_from, tick_from, turn_to, tick_to)
            univbranches = self._universal_cache.settings
            avbranches = self._avatarness_cache.settings
            thbranches = self._things_cache.settings
            rbbranches = self._rulebooks_cache.settings
            trigbranches = self._triggers_cache.settings
            preqbranches = self._prereqs_cache.settings
            actbranches = self._actions_cache.settings
            charrbbranches = self._characters_rulebooks_cache.settings
            avrbbranches = self._avatars_rulebooks_cache.settings
            charthrbbranches = self._characters_things_rulebooks_cache.settings
            charplrbbranches = self._characters_places_rulebooks_cache.settings
            charporbbranches = self._characters_portals_rulebooks_cache.settings
            noderbbranches = self._nodes_rulebooks_cache.settings
            edgerbbranches = self._portals_rulebooks_cache.settings
        else:
            updater = partial(update_backward_window, turn_from, tick_from, turn_to, tick_to)
            univbranches = self._universal_cache.presettings
            avbranches = self._avatarness_cache.presettings
            thbranches = self._things_cache.presettings
            rbbranches = self._rulebooks_cache.presettings
            trigbranches = self._triggers_cache.presettings
            preqbranches = self._prereqs_cache.presettings
            actbranches = self._actions_cache.presettings
            charrbbranches = self._characters_rulebooks_cache.presettings
            avrbbranches = self._avatars_rulebooks_cache.presettings
            charthrbbranches = self._characters_things_rulebooks_cache.presettings
            charplrbbranches = self._characters_places_rulebooks_cache.presettings
            charporbbranches = self._characters_portals_rulebooks_cache.presettings
            noderbbranches = self._nodes_rulebooks_cache.presettings
            edgerbbranches = self._portals_rulebooks_cache.presettings

        def upduniv(_, key, val):
            delta.setdefault('universal', {})[key] = val
        if branch in univbranches:
            updater(upduniv, univbranches[branch])

        def updav(char, graph, node, av):
            delta.setdefault(char, {}).setdefault('avatars', {}).setdefault(graph, {})[node] = bool(av)
        if branch in avbranches:
            updater(updav, avbranches[branch])

        def updthing(char, thing, loc):
            if (
                char in delta and 'nodes' in delta[char]
                and thing in delta[char]['nodes'] and not
                delta[char]['nodes'][thing]
            ):
                return
            thingd = delta.setdefault(char, {}).setdefault('node_val', {}).setdefault(thing, {})
            thingd['location'] = loc
        if branch in thbranches:
            updater(updthing, thbranches[branch])
        # TODO handle arrival_time and next_arrival_time stats of things

        def updrb(whatev, rulebook, rules):
            delta.setdefault('rulebooks', {})[rulebook] = rules

        if branch in rbbranches:
            updater(updrb, rbbranches[branch])

        def updru(key, _, rule, funs):
            delta.setdefault('rules', {}).setdefault(rule, {})[key] = funs

        if branch in trigbranches:
            updater(partial(updru, 'triggers'), trigbranches[branch])

        if branch in preqbranches:
            updater(partial(updru, 'prereqs'), preqbranches[branch])

        if branch in actbranches:
            updater(partial(updru, 'actions'), actbranches[branch])

        def updcrb(key, _, character, rulebook):
            delta.setdefault(character, {})[key] = rulebook

        if branch in charrbbranches:
            updater(partial(updcrb, 'character_rulebook'), charrbbranches[branch])

        if branch in avrbbranches:
            updater(partial(updcrb, 'avatar_rulebook'), avrbbranches[branch])

        if branch in charthrbbranches:
            updater(partial(updcrb, 'character_thing_rulebook'), charthrbbranches[branch])

        if branch in charplrbbranches:
            updater(partial(updcrb, 'character_place_rulebook'), charplrbbranches[branch])

        if branch in charporbbranches:
            updater(partial(updcrb, 'character_portal_rulebook'), charporbbranches[branch])

        def updnoderb(character, node, rulebook):
            if (
                character in delta and 'nodes' in delta[character]
                and node in delta[character]['nodes'] and not delta[character]['nodes'][node]
            ):
                return
            delta.setdefault(character, {}).setdefault('node_val', {}).setdefault(node, {})['rulebook'] = rulebook

        if branch in noderbbranches:
            updater(updnoderb, noderbbranches[branch])

        def updedgerb(character, orig, dest, rulebook):
            if (
                character in delta and 'edges' in delta[character]
                and orig in delta[character]['edges'] and dest in delta[character]['edges'][orig]
                and not delta[character]['edges'][orig][dest]
            ):
                return
            delta.setdefault(character, {}).setdefault('edge_val', {}).setdefault(
                orig, {}).setdefault(dest, {})['rulebook'] = rulebook

        if branch in edgerbbranches:
            updater(updedgerb, edgerbbranches[branch])

        return delta

    def get_turn_delta(self, branch=None, turn=None, tick=None, start_tick=0):
        """Get a dictionary describing changes to the world within a given turn

        Defaults to the present turn, and stops at the present tick unless specified.

        See the documentation for ``get_delta`` for a detailed description of the
        delta format.

        """
        branch = branch or self.branch
        turn = turn or self.turn
        tick = tick or self.tick
        delta = super().get_turn_delta(branch, turn, start_tick, tick)
        if branch in self._avatarness_cache.settings and turn in self._avatarness_cache.settings[branch]:
            for chara, graph, node, is_av in self._avatarness_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(chara, {}).setdefault('avatars', {}).setdefault(graph, {})[node] = is_av
        if branch in self._things_cache.settings and turn in self._things_cache.settings[branch]:
            for chara, thing, location in self._things_cache.settings[branch][turn][start_tick:tick]:
                thingd = delta.setdefault(chara, {}).setdefault('node_val', {}).setdefault(thing, {})
                thingd['location'] = location
        delta['rulebooks'] = rbdif = {}
        if branch in self._rulebooks_cache.settings and turn in self._rulebooks_cache.settings[branch]:
            for _, rulebook, rules in self._rulebooks_cache.settings[branch][turn][start_tick:tick]:
                rbdif[rulebook] = rules
        delta['rules'] = rdif = {}
        if branch in self._triggers_cache.settings and turn in self._triggers_cache.settings[branch]:
            for _, rule, funs in self._triggers_cache.settings[branch][turn][start_tick:tick]:
                rdif.setdefault(rule, {})['triggers'] = funs
        if branch in self._prereqs_cache.settings and turn in self._prereqs_cache.settings[branch]:
            for _, rule, funs in self._prereqs_cache.settings[branch][turn][start_tick:tick]:
                rdif.setdefault(rule, {})['prereqs'] = funs
        if branch in self._actions_cache.settings and turn in self._triggers_cache.settings[branch]:
            for _, rule, funs in self._triggers_cache.settings[branch][turn][start_tick:tick]:
                rdif.setdefault(rule, {})['actions'] = funs

        if branch in self._characters_rulebooks_cache.settings and turn in self._characters_rulebooks_cache.settings[branch]:
            for _, character, rulebook in self._characters_rulebooks_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(character, {})['character_rulebook'] = rulebook
        if branch in self._avatars_rulebooks_cache.settings and turn in self._avatars_rulebooks_cache.settings[branch]:
            for _, character, rulebook in self._avatars_rulebooks_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(character, {})['avatar_rulebook'] = rulebook
        if branch in self._characters_things_rulebooks_cache.settings and turn in self._characters_things_rulebooks_cache.settings[branch]:
            for _, character, rulebook in self._characters_things_rulebooks_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(character, {})['character_thing_rulebook'] = rulebook
        if branch in self._characters_places_rulebooks_cache.settings and turn in self._characters_places_rulebooks_cache.settings[branch]:
            for _, character, rulebook in self._characters_places_rulebooks_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(character, {})['character_place_rulebook'] = rulebook
        if branch in self._characters_portals_rulebooks_cache.settings and turn in self._characters_portals_rulebooks_cache.settings[branch]:
            for _, character, rulebook in self._characters_portals_rulebooks_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(character, {})['character_portal_rulebook'] = rulebook

        if branch in self._nodes_rulebooks_cache.settings and turn in self._nodes_rulebooks_cache.settings[branch]:
            for character, node, rulebook in self._nodes_rulebooks_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(character, {}).setdefault('node_val', {}).setdefault(node, {})['rulebook'] = rulebook
        if branch in self._portals_rulebooks_cache.settings and turn in self._portals_rulebooks_cache.settings[branch]:
            for character, orig, dest, rulebook in self._portals_rulebooks_cache.settings[branch][turn][start_tick:tick]:
                delta.setdefault(character, {}).setdefault('edge_val', {})\
                    .setdefault(orig, {}).setdefault(dest, {})['rulebook'] = rulebook
        return delta

    def _del_rulebook(self, rulebook):
        for (character, character_rulebooks) in \
             self._characters_rulebooks_cache.items():
            if rulebook not in character_rulebooks.values():
                continue
            for (which, rb) in character_rulebooks.items():
                if rb == rulebook:
                    raise ValueError(
                        "Rulebook still in use by {} as {}".format(
                            character, which
                        ))
        for (character, nodes) in self._nodes_rulebooks_cache.items():
            if rulebook not in nodes.values():
                continue
            for (node, rb) in nodes.items():
                if rb == rulebook:
                    raise ValueError(
                        "Rulebook still in use by node "
                        "{} in character {}".format(
                            node, character
                        ))
        for (character, origins) in self._portals_rulebooks_cache.items():
            for (origin, destinations) in origins.items():
                if rulebook not in destinations.values():
                    continue
                for (destination, rb) in destinations:
                    if rb == rulebook:
                        raise ValueError(
                            "Rulebook still in use by portal "
                            "{}->{} in character {}".format(
                                origin, destination, character
                            ))
        self.rule.query.rulebook_del_all(rulebook)
        del self._rulebooks_cache._data[rulebook]

    def _remember_avatarness(
            self, character, graph, node,
            is_avatar=True, branch=None, turn=None,
            tick=None
    ):
        """Use this to record a change in avatarness.

        Should be called whenever a node that wasn't an avatar of a
        character now is, and whenever a node that was an avatar of a
        character now isn't.

        ``character`` is the one using the node as an avatar,
        ``graph`` is the character the node is in.

        """
        branch = branch or self.branch
        turn = turn or self.turn
        tick = tick or self.tick
        self._avatarness_cache.store(
            character,
            graph,
            node,
            branch,
            turn,
            tick,
            is_avatar
        )
        self.query.avatar_set(
            character,
            graph,
            node,
            branch,
            turn,
            tick,
            is_avatar
        )

    def _init_caches(self):
        from .xcollections import (
            StringStore,
            FunctionStore,
            CharacterMapping,
            UniversalMapping
        )
        from .cache import (
            Cache,
            InitializedCache,
            EntitylessCache,
            InitializedEntitylessCache,
            AvatarnessCache,
            AvatarRulesHandledCache,
            CharacterThingRulesHandledCache,
            CharacterPlaceRulesHandledCache,
            CharacterPortalRulesHandledCache,
            NodeRulesHandledCache,
            PortalRulesHandledCache,
            CharacterRulesHandledCache,
            ThingsCache
        )
        from .rule import AllRuleBooks, AllRules

        super()._init_caches()
        self._things_cache = ThingsCache(self)
        self._node_contents_cache = Cache(self)
        self.character = self.graph = CharacterMapping(self)
        self._universal_cache = EntitylessCache(self)
        self._rulebooks_cache = InitializedEntitylessCache(self)
        self._characters_rulebooks_cache = InitializedEntitylessCache(self)
        self._avatars_rulebooks_cache = InitializedEntitylessCache(self)
        self._characters_things_rulebooks_cache = InitializedEntitylessCache(self)
        self._characters_places_rulebooks_cache = InitializedEntitylessCache(self)
        self._characters_portals_rulebooks_cache = InitializedEntitylessCache(self)
        self._nodes_rulebooks_cache = InitializedCache(self)
        self._portals_rulebooks_cache = InitializedCache(self)
        self._triggers_cache = InitializedEntitylessCache(self)
        self._prereqs_cache = InitializedEntitylessCache(self)
        self._actions_cache = InitializedEntitylessCache(self)
        self._node_rules_handled_cache = NodeRulesHandledCache(self)
        self._portal_rules_handled_cache = PortalRulesHandledCache(self)
        self._character_rules_handled_cache = CharacterRulesHandledCache(self)
        self._avatar_rules_handled_cache = AvatarRulesHandledCache(self)
        self._character_thing_rules_handled_cache \
            = CharacterThingRulesHandledCache(self)
        self._character_place_rules_handled_cache \
            = CharacterPlaceRulesHandledCache(self)
        self._character_portal_rules_handled_cache \
            = CharacterPortalRulesHandledCache(self)
        self._avatarness_cache = AvatarnessCache(self)
        self._turns_completed = defaultdict(lambda: max((0, self.turn - 1)))
        """The last turn when the rules engine ran in each branch"""
        self.eternal = self.query.globl
        self.universal = UniversalMapping(self)
        if hasattr(self, '_action_file'):
            self.action = FunctionStore(self._action_file)
        if hasattr(self, '_prereq_file'):
            self.prereq = FunctionStore(self._prereq_file)
        if hasattr(self, '_trigger_file'):
            self.trigger = FunctionStore(self._trigger_file)
        if hasattr(self, '_function_file'):
            self.function = FunctionStore(self._function_file)
        if hasattr(self, '_method_file'):
            self.method = FunctionStore(self._method_file)
        self.rule = AllRules(self)
        self.rulebook = AllRuleBooks(self)
        if hasattr(self, '_string_file'):
            self.string = StringStore(
                self.query,
                self._string_file,
                self.eternal.setdefault('language', 'eng')
            )

    def _load_graphs(self):
        for charn in self.query.characters():
            self._graph_objs[charn] = self.char_cls(self, charn, init_rulebooks=False)

    def __init__(
            self,
            worlddb,
            *,
            string='strings.json',
            function='function.py',
            method='method.py',
            trigger='trigger.py',
            prereq='prereq.py',
            action='action.py',
            connect_args={},
            alchemy=False,
            commit_modulus=None,
            random_seed=None,
            logfun=None,
            validate=False,
            clear_code=False,
            clear_world=False
    ):
        """Store the connections for the world database and the code database;
        set up listeners; and start a transaction

        """
        import os
        worlddbpath = worlddb.replace('sqlite:///', '')
        if clear_world and os.path.exists(worlddbpath):
            os.remove(worlddbpath)
        if isinstance(string, str):
            self._string_file = string
            if clear_code and os.path.exists(string):
                os.remove(string)
        else:
            self.string = string
        if isinstance(function, str):
            self._function_file = function
            if clear_code and os.path.exists(function):
                os.remove(function)
        else:
            self.function = function
        if isinstance(method, str):
            self._method_file = method
            if clear_code and os.path.exists(method):
                os.remove(method)
        else:
            self.method = method
        if isinstance(trigger, str):
            self._trigger_file = trigger
            if clear_code and os.path.exists(trigger):
                os.remove(trigger)
        else:
            self.trigger = trigger
        if isinstance(prereq, str):
            self._prereq_file = prereq
            if clear_code and os.path.exists(prereq):
                os.remove(prereq)
        else:
            self.prereq = prereq
        if isinstance(action, str):
            self._action_file = action
            if clear_code and os.path.exists(action):
                os.remove(action)
        else:
            self.action = action
        super().__init__(
            worlddb,
            connect_args=connect_args,
            alchemy=alchemy,
            validate=validate
        )
        self.next_turn = NextTurn(self)
        if logfun is None:
            from logging import getLogger
            logger = getLogger(__name__)

            def logfun(level, msg):
                getattr(logger, level)(msg)
        self.log = logfun
        self.commit_modulus = commit_modulus
        self.random_seed = random_seed
        self._rules_iter = self._follow_rules()
        # set up the randomizer
        from random import Random
        self._rando = Random()
        if 'rando_state' in self.universal:
            self._rando.setstate(self.universal['rando_state'])
        else:
            self._rando.seed(self.random_seed)
            self.universal['rando_state'] = self._rando.getstate()
        if hasattr(self.method, 'init'):
            self.method.init(self)

    def _init_load(self, validate=False):
        from .rule import Rule
        q = self.query
        self._things_cache.load(q.things_dump(), validate)
        super()._init_load(validate=validate)
        self._avatarness_cache.load(q.avatars_dump(), validate)
        self._universal_cache.load(q.universals_dump(), validate)
        self._rulebooks_cache.load(q.rulebooks_dump(), validate)
        self._characters_rulebooks_cache.load(q.character_rulebook_dump(), validate)
        self._avatars_rulebooks_cache.load(q.avatar_rulebook_dump(), validate)
        self._characters_things_rulebooks_cache.load(q.character_thing_rulebook_dump(), validate)
        self._characters_places_rulebooks_cache.load(q.character_place_rulebook_dump(), validate)
        self._characters_portals_rulebooks_cache.load(q.character_portal_rulebook_dump(), validate)
        self._nodes_rulebooks_cache.load(q.node_rulebook_dump(), validate)
        self._portals_rulebooks_cache.load(q.portal_rulebook_dump(), validate)
        self._triggers_cache.load(q.rule_triggers_dump(), validate)
        self._prereqs_cache.load(q.rule_prereqs_dump(), validate)
        self._actions_cache.load(q.rule_actions_dump(), validate)
        for row in q.character_rules_handled_dump():
            self._character_rules_handled_cache.store(*row, loading=True)
        for row in q.avatar_rules_handled_dump():
            self._avatar_rules_handled_cache.store(*row, loading=True)
        for row in q.character_thing_rules_handled_dump():
            self._character_thing_rules_handled_cache.store(*row, loading=True)
        for row in q.character_place_rules_handled_dump():
            self._character_place_rules_handled_cache.store(*row, loading=True)
        for row in q.character_portal_rules_handled_dump():
            self._character_portal_rules_handled_cache.store(*row, loading=True)
        for row in q.node_rules_handled_dump():
            self._node_rules_handled_cache.store(*row, loading=True)
        for row in q.portal_rules_handled_dump():
            self._portal_rules_handled_cache.store(*row, loading=True)
        self._turns_completed.update(q.turns_completed_dump())
        self._rules_cache = {name: Rule(self, name, create=False) for name in q.rules_dump()}

    @property
    def stores(self):
        return (
            self.action,
            self.prereq,
            self.trigger,
            self.function,
            self.method,
            self.string
        )

    def debug(self, msg):
        """Log a message at level 'debug'"""
        self.log('debug', msg)

    def info(self, msg):
        """Log a message at level 'info'"""
        self.log('info', msg)

    def warning(self, msg):
        """Log a message at level 'warning'"""
        self.log('warning', msg)

    def error(self, msg):
        """Log a message at level 'error'"""
        self.log('error', msg)

    def critical(self, msg):
        """Log a message at level 'critical'"""
        self.log('critical', msg)

    def close(self):
        """Commit changes and close the database."""
        for store in self.stores:
            if hasattr(store, 'save'):
                store.save()
        super().close()

    def __enter__(self):
        """Return myself. For compatibility with ``with`` semantics."""
        return self

    def __exit__(self, *args):
        """Close on exit."""
        self.close()

    def _set_branch(self, v):
        super()._set_branch(v)
        self.time.send(self.time, branch=self._obranch, turn=self._oturn)

    def _set_turn(self, v):
        super()._set_turn(v)
        self.time.send(self.time, branch=self._obranch, turn=self._oturn)

    def _handled_char(self, charn, rulebook, rulen, branch, turn, tick):
        try:
            self._character_rules_handled_cache.store(
                charn, rulebook, rulen, branch, turn, tick
            )
        except ValueError:
            assert rulen in self._character_rules_handled_cache.handled[
                charn, rulebook, branch, turn
            ]
            return
        self.query.handled_character_rule(
            charn, rulebook, rulen, branch, turn, tick
        )

    def _handled_av(self, character, graph, avatar, rulebook, rule, branch, turn, tick):
        try:
            self._avatar_rules_handled_cache.store(
                character, graph, avatar, rulebook, rule, branch, turn, tick
            )
        except ValueError:
            assert rule in self._avatar_rules_handled_cache.handled[
                character, graph, avatar, rulebook, branch, turn
            ]
            return
        self.query.handled_avatar_rule(
            character, rulebook, rule, graph, avatar, branch, turn, tick
        )

    def _handled_char_thing(self, character, thing, rulebook, rule, branch, turn, tick):
        try:
            self._character_thing_rules_handled_cache.store(
                character, thing, rulebook, rule, branch, turn, tick
            )
        except ValueError:
            assert rule in self._character_thing_rules_handled_cache.handled[
                character, thing, rulebook, branch, turn
            ]
            return
        self.query.handled_character_thing_rule(
            character, rulebook, rule, thing, branch, turn, tick
        )

    def _handled_char_place(self, character, place, rulebook, rule, branch, turn, tick):
        try:
            self._character_place_rules_handled_cache.store(
                character, place, rulebook, rule, branch, turn, tick
            )
        except ValueError:
            assert rule in self._character_place_rules_handled_cache.handled[
                character, place, rulebook, branch, turn
            ]
            return
        self.query.handled_character_place_rule(
            character, rulebook, rule, place, branch, turn, tick
        )

    def _handled_char_port(self, character, orig, dest, rulebook, rule, branch, turn, tick):
        try:
            self._character_portal_rules_handled_cache.store(
                character, orig, dest, rulebook, rule, branch, turn, tick
            )
        except ValueError:
            assert rule in self._character_portal_rules_handled_cache.handled[
                character, orig, dest, rulebook, branch, turn
            ]
            return
        self.query.handled_character_portal_rule(
            character, orig, dest, rulebook, rule, branch, turn, tick
        )

    def _handled_node(self, character, node, rulebook, rule, branch, turn, tick):
        try:
            self._node_rules_handled_cache.store(
                character, node, rulebook, rule, branch, turn, tick
            )
        except ValueError:
            assert rule in self._node_rules_handled_cache.handled[
                character, node, rulebook, branch, turn
            ]
            return
        self.query.handled_node_rule(
            character, node, rulebook, rule, branch, turn, tick
        )

    def _handled_portal(self, character, orig, dest, rulebook, rule, branch, turn, tick):
        try:
            self._portal_rules_handled_cache.store(
                character, orig, dest, rulebook, rule, branch, turn, tick
            )
        except ValueError:
            assert rule in self._portal_rules_handled_cache.handled[
                character, orig, dest, rulebook, branch, turn
            ]
            return
        self.query.handled_portal_rule(
            character, orig, dest, rulebook, rule, branch, turn, tick
        )

    def _follow_rule(self, rule, handled_fun, branch, turn, *args):
        self.debug("following rule: " + repr(rule))
        satisfied = True
        for prereq in rule.prereqs:
            res = prereq(*args)
            if not res:
                satisfied = False
                break
        if not satisfied:
            return handled_fun()
        for trigger in rule.triggers:
            res = trigger(*args)
            if res:
                break
        else:
            return handled_fun()
        actres = []
        for action in rule.actions:
            res = action(*args)
            if res:
                actres.append(res)
        handled_fun()
        return actres

    def _follow_rules(self):
        # TODO: roll back changes done by rules that raise an exception
        # TODO: if there's a paradox while following some rule, start a new branch, copying handled rules
        from collections import defaultdict
        branch, turn, tick = self.btt()
        charmap = self.character
        rulemap = self.rule
        todo = defaultdict(list)

        def do_rule(tup):
            # Returns None if the entity following the rule no longer exists.
            # Better way to handle this?
            return {
                'character': lambda charactername, rulebook, rulename: self._follow_rule(
                    rulemap[rulename],
                    partial(self._handled_char, charactername, rulebook, rulename, branch, turn, tick),
                    branch, turn,
                    charmap[charactername]
                ) if charactername in charmap else None,
                'avatar': lambda charn, rulebook, graphn, avn, rulen: self._follow_rule(
                    rulemap[rulen],
                    partial(self._handled_av, charn, graphn, avn, rulebook, rulen, branch, turn, tick),
                    branch, turn,
                    charmap[graphn].node[avn]
                ) if self._node_exists(graphn, avn) else None,
                'character_thing': lambda charn, rulebook, rulen, thingn: self._follow_rule(
                    rulemap[rulen],
                    partial(self._handled_char_thing, charn, thingn, rulebook, rulen, branch, turn, tick),
                    branch, turn,
                    charmap[charn].thing[thingn]
                ) if charn in charmap and thingn in charmap[charn].thing else None,
                'character_place': lambda charn, rulebook, rulen, placen: self._follow_rule(
                    rulemap[rulen],
                    partial(self._handled_char_place, charn, placen, rulebook, rulen, branch, turn, tick),
                    branch, turn,
                    charmap[charn].place[placen]
                ) if charn in charmap and placen in charmap[charn].place else None,
                'character_portal': lambda charn, rulebook, rulen, orign, destn: self._follow_rule(
                    rulemap[rulen],
                    partial(self._handled_char_port, charn, orign, destn, rulebook, rulen, branch, turn, tick),
                    branch, turn,
                    charmap[charn].portal[orign][destn]
                ) if self._edge_exists(charn, orign, destn) else None,
                'node': lambda charn, noden, rulebook, rulen: self._follow_rule(
                    rulemap[rulen],
                    partial(self._handled_node, charn, noden, rulebook, rulen, branch, turn, tick),
                    branch, turn,
                    charmap[charn].node[noden]
                ) if self._node_exists(charn, noden) else None,
                'portal': lambda charn, orign, destn, rulebook, rulen: self._follow_rule(
                    rulemap[rulen],
                    partial(self._handled_portal, charn, orign, destn, rulebook, rulen, branch, turn, tick),
                    branch, turn,
                    charmap[charn].portal[orign][destn]
                ) if self._edge_exists(charn, orign, destn) else None
            }[tup[0]](*tup[1:])
        for (
            charactername, rulebook, rulename
        ) in self._character_rules_handled_cache.iter_unhandled_rules(
                branch, turn, tick
        ):
            if charactername not in charmap:
                continue
            todo[rulebook].append(('character', charactername, rulebook, rulename))
        for (
            charn, graphn, avn, rulebook, rulen
        ) in self._avatar_rules_handled_cache.iter_unhandled_rules(
                branch, turn, tick
        ):
            if charn not in charmap:
                continue
            char = charmap[charn]
            if graphn not in char.avatar or avn not in char.avatar[graphn]:
                continue
            todo[rulebook].append(('avatar', charn, rulebook, graphn, avn, rulen))
        for (
            charn, thingn, rulebook, rulen
        ) in self._character_thing_rules_handled_cache.iter_unhandled_rules(branch, turn, tick):
            if charn not in charmap or thingn not in charmap[charn].thing:
                continue
            todo[rulebook].append(('character_thing', charn, rulebook, rulen, thingn))
        for (
            charn, placen, rulebook, rulen
        ) in self._character_place_rules_handled_cache.iter_unhandled_rules(
            branch, turn, tick
        ):
            if charn not in charmap or placen not in charmap[charn].place:
                continue
            todo[rulebook].append(('character_place', charn, rulebook, rulen, placen))
        for (
            charn, orign, destn, rulebook, rulen
        ) in self._character_portal_rules_handled_cache.iter_unhandled_rules(
            branch, turn, tick
        ):
            if charn not in charmap:
                continue
            char = charmap[charn]
            if orign not in char.portal or destn not in char.portal[orign]:
                continue
            todo[rulebook].append(('character_portal', charn, rulebook, rulen, orign, destn))
        for (
                charn, noden, rulebook, rulen
        ) in self._node_rules_handled_cache.iter_unhandled_rules(
            branch, turn, tick
        ):
            if charn not in charmap or noden not in charmap[charn]:
                continue
            todo[rulebook].append(('node', charn, noden, rulebook, rulen))
        for (
                charn, orign, destn, rulebook, rulen
        ) in self._portal_rules_handled_cache.iter_unhandled_rules(
                branch, turn, tick
        ):
            if charn not in charmap:
                continue
            char = charmap[charn]
            if orign not in char.portal or destn not in char.portal[orign]:
                continue
            todo[rulebook].append(('portal', charn, orign, destn, rulebook, rulen))

        # TODO: rulebook priorities (not individual rule priorities, just follow the order of the rulebook)
        for rulebook in sort_set(todo.keys()):
            for rule in todo[rulebook]:
                try:
                    yield do_rule(rule)
                except StopIteration:
                    raise InnerStopIteration

    def advance(self):
        """Follow the next rule if available.

        If we've run out of rules, reset the rules iterator.

        """
        try:
            return next(self._rules_iter)
        except InnerStopIteration:
            self._rules_iter = self._follow_rules()
            return StopIteration()
        except StopIteration:
            self._rules_iter = self._follow_rules()
            return final_rule
        # except Exception as ex:
        #     self._rules_iter = self._follow_rules()
        #     return ex

    def new_character(self, name, data=None, **kwargs):
        """Create and return a new :class:`Character`."""
        self.add_character(name, data, **kwargs)
        return self.character[name]

    def add_character(self, name, data=None, **kwargs):
        """Create a new character.

        You'll be able to access it as a :class:`Character` object by
        looking up ``name`` in my ``character`` property.

        ``data``, if provided, should be a networkx-compatible graph
        object. Your new character will be a copy of it.

        Any keyword arguments will be set as stats of the new character.

        """
        self._init_graph(name, 'DiGraph')
        self._graph_objs[name] = self.char_cls(self, name, data, **kwargs)

    def del_character(self, name):
        """Remove the Character from the database entirely.

        This also deletes all its history. You'd better be sure.

        """
        self.query.del_character(name)
        self.del_graph(name)
        del self.character[name]

    def _is_thing(self, character, node):
        return self._things_cache.contains_entity(character, node, *self.btt())

    def _set_thing_loc(
            self, character, node, loc
    ):
        branch, turn, tick = self.nbtt()
        self._things_cache.store(character, node, branch, turn, tick, loc)
        self.query.set_thing_loc(
            character,
            node,
            branch,
            turn,
            tick,
            loc
        )

    def alias(self, v, stat='dummy'):
        from .util import EntityStatAccessor
        r = DummyEntity(self)
        r[stat] = v
        return EntityStatAccessor(r, stat, engine=self)

    def entityfy(self, v, stat='dummy'):
        from .query import Query
        from .util import EntityStatAccessor
        if (
                isinstance(v, self.thing_cls) or
                isinstance(v, self.place_cls) or
                isinstance(v, self.portal_cls) or
                isinstance(v, Query) or
                isinstance(v, EntityStatAccessor)
        ):
            return v
        return self.alias(v, stat)

    def turns_when(self, qry):
        for branch, turn in qry.iter_turns():
            yield turn


# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Common utility functions and data structures.

"""
from collections.abc import Set
from operator import attrgetter, add, sub, mul, pow, truediv, floordiv, mod
from functools import partial
from textwrap import dedent
from .reify import reify


def getatt(attribute_name):
    """An easy way to make an alias"""
    return property(attrgetter(attribute_name))


def singleton_get(s):
    """Take an iterable and return its only item if possible, else None."""
    it = None
    for that in s:
        if it is not None:
            return None
        it = that
    return it


class EntityStatAccessor(object):
    __slots__ = [
        'engine', 'entity', 'branch', 'turn', 'tick', 'stat', 'current', 'mungers'
    ]

    def __init__(
            self, entity, stat,
            engine=None, branch=None, turn=None, tick=None, current=False, mungers=[]
    ):
        if engine is None:
            engine = entity.engine
        if branch is None:
            branch = engine.branch
        if turn is None:
            turn = engine.turn
        if tick is None:
            tick = engine.tick
        self.current = current
        self.engine = engine
        self.entity = entity
        self.stat = stat
        self.branch = branch
        self.turn = turn
        self.tick = tick
        self.mungers = mungers

    def __call__(self, branch=None, turn=None, tick=None):
        if self.current:
            res = self.entity[self.stat]
        else:
            branc, trn, tck = self.engine.btt()
            self.engine.branch = branch or self.branch
            self.engine.turn = turn or self.turn
            self.engine.tick = tick or self.tick
            res = self.entity[self.stat]
            self.engine.branch = branc
            self.engine.turn = trn
            self.engine.tick = tck
        for munger in self.mungers:
            res = munger(res)
        return res

    def __ne__(self, other):
        return self() != other

    def __str__(self):
        return str(self())

    def __repr__(self):
        return "EntityStatAccessor({}[{}]{}), {} mungers".format(
            self.entity,
            self.stat,
            "" if self.current else
            ", branch={}, turn={}, tick={}".format(self.branch, self.turn, self.tick),
            len(self.mungers)
        )

    def __gt__(self, other):
        return self() > other

    def __ge__(self, other):
        return self >= other

    def __lt__(self, other):
        return self < other

    def __le__(self, other):
        return self <= other

    def __eq__(self, other):
        return self == other

    def munge(self, munger):
        return EntityStatAccessor(
            self.entity,
            self.stat,
            self.engine,
            self.branch,
            self.tick,
            self.current,
            self.mungers + [munger]
        )

    def __add__(self, other):
        return self.munge(partial(add, other))

    def __sub__(self, other):
        return self.munge(partial(sub, other))

    def __mul__(self, other):
        return self.munge(partial(mul, other))

    def __rpow__(self, other, modulo=None):
        return self.munge(partial(pow, other, modulo=modulo))

    def __rdiv__(self, other):
        return self.munge(partial(truediv, other))

    def __rfloordiv__(self, other):
        return self.munge(partial(floordiv, other))

    def __rmod__(self, other):
        return self.munge(partial(mod, other))

    def __getitem__(self, k):
        return self.munge(lambda x: x[k])


def dedent_source(source):
    nlidx = source.index('\n')
    if nlidx is None:
        raise ValueError("Invalid source")
    while source[:nlidx].strip().startswith('@'):
        source = source[nlidx+1:]
        nlidx = source.index('\n')
    return dedent(source)


def _sort_set_key(v):
    if isinstance(v, tuple):
        return (2,) + tuple(map(repr, v))
    if isinstance(v, str):
        return 1, repr(v)
    return 0, repr(v)


_sort_set_memo = {}


def sort_set(s):
    """Return a sorted list of the contents of a set

    This is intended to be used to iterate over world state, where you just need keys
    to be in some deterministic order, but the sort order should be obvious from the key.

    Non-strings come before strings and then tuples. Tuples compare element-wise as normal.
    But ultimately all comparisons are between values' ``repr``.

    This is memoized.

    """
    if not isinstance(s, Set):
        raise TypeError("sets only")
    s = frozenset(s)
    if s not in _sort_set_memo:
        _sort_set_memo[s] = sorted(s, key=_sort_set_key)
    return _sort_set_memo[s]



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The query engine provides Pythonic methods to access the database.

This module also contains a notably unfinished implementation of a query
language specific to LiSE. Access some stats using entities' method
``historical``, and do comparisons on those, and instead of a boolean
result you'll get a callable object that will return an iterator over
turn numbers in which the comparison evaluated to ``True``.

"""
from operator import gt, lt, eq, ne, le, ge
from functools import partialmethod

import allegedb.query

from .exc import (
    IntegrityError,
    OperationalError
)
from .util import EntityStatAccessor
import LiSE


def windows_union(windows):
    """Given a list of (beginning, ending), return a minimal version that contains the same ranges.

    :rtype: list

    """
    def fix_overlap(left, right):
        if left == right:
            return [left]
        assert left[0] < right[0]
        if left[1] >= right[0]:
            if right[1] > left[1]:
                return [(left[0], right[1])]
            else:
                return [left]
        return [left, right]

    if len(windows) == 1:
        return windows
    none_left = []
    none_right = []
    otherwise = []
    for window in windows:
        if window[0] is None:
            none_left.append(window)
        elif window[1] is None:
            none_right.append(window)
        else:
            otherwise.append(window)

    res = []
    otherwise.sort()
    for window in none_left:
        if not res:
            res.append(window)
            continue
        res.extend(fix_overlap(res.pop(), window))
    while otherwise:
        window = otherwise.pop(0)
        if not res:
            res.append(window)
            continue
        res.extend(fix_overlap(res.pop(), window))
    for window in none_right:
        if not res:
            res.append(window)
            continue
        res.extend(fix_overlap(res.pop(), window))
    return res


def windows_intersection(windows):
    """Given a list of (beginning, ending), return another describing where they overlap.

    :rtype: list
    """

    def intersect2(left, right):
        if left == right:
            return left
        elif left is (None, None):
            return right
        elif right is (None, None):
            return left
        elif left[0] is None:
            if right[0] is None:
                return None, min((left[1], right[1]))
            elif right[1] is None:
                if left[1] <= right[0]:
                    return left[1], right[0]
                else:
                    return None
            elif right[0] <= left[1]:
                return right[0], left[1]
            else:
                return None
        elif left[1] is None:
            if right[0] is None:
                return left[0], right[1]
            else:
                return right  # assumes left[0] <= right[0]
        # None not in left
        elif right[0] is None:
            return left[0], min((left[1], right[1]))
        elif right[1] is None:
            if left[1] >= right[0]:
                return right[0], left[1]
            else:
                return None
        assert None not in left and None not in right and left[0] < right[1]
        if left[1] >= right[0]:
            if right[1] > left[1]:
                return right[0], left[1]
            else:
                return right
        return None

    if len(windows) == 1:
        return windows
    left_none = []
    right_none = []
    otherwise = []
    for window in windows:
        assert window is not None, None
        if window[0] is None:
            left_none.append(window)
        elif window[1] is None:
            right_none.append(window)
        else:
            otherwise.append(window)

    done = []
    todo = left_none + sorted(otherwise)
    for window in todo:
        if not done:
            done.append(window)
            continue
        res = intersect2(done.pop(), window)
        if res:
            done.append(res)
    return done


class Query(object):
    def __new__(cls, engine, leftside, rightside=None, **kwargs):
        if rightside is None:
            if not isinstance(leftside, cls):
                raise TypeError("You can't make a query with only one side")
            me = leftside
        else:
            me = super().__new__(cls)
            me.leftside = leftside
            me.rightside = rightside
        me.engine = engine
        me.windows = kwargs.get('windows', [])
        return me

    def iter_turns(self):
        raise NotImplementedError

    def iter_ticks(self, turn):
        raise NotImplementedError

    def __eq__(self, other):
        return EqQuery(self.engine, self, self.engine.entityfy(other))

    def __gt__(self, other):
        return GtQuery(self.engine, self, self.engine.entityfy(other))

    def __ge__(self, other):
        return GeQuery(self.engine, self, self.engine.entityfy(other))

    def __lt__(self, other):
        return LtQuery(self.engine, self, self.engine.entityfy(other))

    def __le__(self, other):
        return LeQuery(self.engine, self, self.engine.entityfy(other))

    def __ne__(self, other):
        return NeQuery(self.engine, self, self.engine.entityfy(other))

    def and_before(self, end):
        if self.windows:
            new_windows = windows_intersection(
                sorted(self.windows + [(None, end)])
            )
        else:
            new_windows = [(0, end)]
        return type(self)(self.leftside, self.rightside, windows=new_windows)
    before = and_before

    def or_before(self, end):
        if self.windows:
            new_windows = windows_union(self.windows + [(None, end)])
        else:
            new_windows = [(None, end)]
        return type(self)(self.leftside, self.rightside, windows=new_windows)

    def and_after(self, start):
        if self.windows:
            new_windows = windows_intersection(self.windows + [(start, None)])
        else:
            new_windows = [(start, None)]
        return type(self)(self.leftside, self.rightside, windows=new_windows)
    after = and_after

    def or_between(self, start, end):
        if self.windows:
            new_windows = windows_union(self.windows + [(start, end)])
        else:
            new_windows = [(start, end)]
        return type(self)(self.leftside, self.rightside, windows=new_windows)

    def and_between(self, start, end):
        if self.windows:
            new_windows = windows_intersection(self.windows + [(start, end)])
        else:
            new_windows = [(start, end)]
        return type(self)(self.leftside, self.rightside, windows=new_windows)
    between = and_between

    def or_during(self, tick):
        return self.or_between(tick, tick)

    def and_during(self, tick):
        return self.and_between(tick, tick)
    during = and_during


class Union(Query):
    pass


class ComparisonQuery(Query):
    oper = lambda x, y: NotImplemented

    def iter_turns(self):
        return slow_iter_turns_eval_cmp(self, self.oper, engine=self.engine)


class EqQuery(ComparisonQuery):
    oper = eq


class NeQuery(ComparisonQuery):
    oper = ne


class GtQuery(ComparisonQuery):
    oper = gt


class LtQuery(ComparisonQuery):
    oper = lt


class GeQuery(ComparisonQuery):
    oper = ge


class LeQuery(ComparisonQuery):
    oper = le


comparisons = {
    'eq': EqQuery,
    'ne': NeQuery,
    'gt': GtQuery,
    'lt': LtQuery,
    'ge': GeQuery,
    'le': LeQuery
}


class StatusAlias(EntityStatAccessor):
    def __eq__(self, other):
        return EqQuery(self.engine, self, other)

    def __ne__(self, other):
        return NeQuery(self.engine, self, other)

    def __gt__(self, other):
        return GtQuery(self.engine, self, other)

    def __lt__(self, other):
        return LtQuery(self.engine, self, other)

    def __ge__(self, other):
        return GeQuery(self.engine, self, other)

    def __le__(self, other):
        return LeQuery(self.engine, self, other)


def slow_iter_turns_eval_cmp(qry, oper, start_branch=None, engine=None):
    """Iterate over all turns on which a comparison holds.

    This is expensive. It evaluates the query for every turn in history.

    """
    def mungeside(side):
        if isinstance(side, Query):
            return side.iter_turns
        elif isinstance(side, StatusAlias):
            return EntityStatAccessor(
                side.entity, side.stat, side.engine,
                side.branch, side.turn, side.tick, side.current, side.mungers
            )
        elif isinstance(side, EntityStatAccessor):
            return side
        else:
            return lambda: side
    leftside = mungeside(qry.leftside)
    rightside = mungeside(qry.rightside)
    engine = engine or leftside.engine or rightside.engine

    for (branch, _, _) in engine._iter_parent_btt(start_branch or engine.branch):
        if branch is None:
            return
        parent, turn_start, tick_start, turn_end, tick_end = engine._branches[branch]
        for turn in range(turn_start, engine.turn + 1):
            if oper(leftside(branch, turn), rightside(branch, turn)):
                yield branch, turn


class QueryEngine(allegedb.query.QueryEngine):
    path = LiSE.__path__[0]
    IntegrityError = IntegrityError
    OperationalError = OperationalError

    def universals_dump(self):
        for key, branch, turn, tick, value in self.sql('universals_dump'):
            yield self.unpack(key), branch, turn, tick, self.unpack(value)

    def rulebooks_dump(self):
        for rulebook, branch, turn, tick, rules in self.sql('rulebooks_dump'):
            yield self.unpack(rulebook), branch, turn, tick, self.unpack(rules)

    def _rule_dump(self, typ):
        for rule, branch, turn, tick, lst in self.sql('rule_{}_dump'.format(typ)):
            yield rule, branch, turn, tick, self.unpack(lst)

    def rule_triggers_dump(self):
        return self._rule_dump('triggers')

    def rule_prereqs_dump(self):
        return self._rule_dump('prereqs')

    def rule_actions_dump(self):
        return self._rule_dump('actions')

    def characters_dump(self):
        for graph, typ in self.sql('graphs_dump'):
            if typ == 'DiGraph':
                yield self.unpack(graph)
    characters = characters_dump

    def node_rulebook_dump(self):
        for character, node, branch, turn, tick, rulebook in self.sql('node_rulebook_dump'):
            yield self.unpack(character), self.unpack(node), branch, turn, tick, self.unpack(rulebook)

    def portal_rulebook_dump(self):
        for character, orig, dest, branch, turn, tick, rulebook in self.sql('portal_rulebook_dump'):
            yield (
                self.unpack(character), self.unpack(orig), self.unpack(dest),
                branch, turn, tick, self.unpack(rulebook)
            )

    def _charactery_rulebook_dump(self, qry):
        for character, branch, turn, tick, rulebook in self.sql(qry+'_rulebook_dump'):
            yield self.unpack(character), branch, turn, tick, self.unpack(rulebook)

    character_rulebook_dump = partialmethod(_charactery_rulebook_dump, 'character')
    avatar_rulebook_dump = partialmethod(_charactery_rulebook_dump, 'avatar')
    character_thing_rulebook_dump = partialmethod(_charactery_rulebook_dump, 'character_thing')
    character_place_rulebook_dump = partialmethod(_charactery_rulebook_dump, 'character_place')
    character_portal_rulebook_dump = partialmethod(_charactery_rulebook_dump, 'character_portal')

    def character_rules_handled_dump(self):
        for character, rulebook, rule, branch, turn, tick in self.sql('character_rules_handled_dump'):
            yield self.unpack(character), self.unpack(rulebook), rule, branch, turn, tick

    def character_rules_changes_dump(self):
        for (
                character, rulebook, rule, branch, turn, tick, handled_branch, handled_turn
        ) in self.sql('character_rules_changes_dump'):
            yield (
                self.unpack(character), self.unpack(rulebook),
                rule, branch, turn, tick, handled_branch, handled_turn
            )

    def avatar_rules_handled_dump(self):
        for character, rulebook, rule, graph, avatar, branch, turn, tick in self.sql('avatar_rules_handled_dump'):
            yield (
                self.unpack(character), self.unpack(rulebook), rule,
                self.unpack(graph), self.unpack(avatar), branch, turn, tick
            )

    def avatar_rules_changes_dump(self):
        jl = self.unpack
        for (
            character, rulebook, rule, graph, avatar, branch, turn, tick, handled_branch, handled_turn
        ) in self.sql('avatar_rules_changes_dump'):
            yield (
                jl(character), jl(rulebook), rule, jl(graph), jl(avatar),
                branch, turn, tick, handled_branch, handled_turn
            )

    def character_thing_rules_handled_dump(self):
        for character, rulebook, rule, thing, branch, turn, tick in self.sql('character_thing_rules_handled_dump'):
            yield self.unpack(character), self.unpack(rulebook), rule, self.unpack(thing), branch, turn, tick

    def character_thing_rules_changes_dump(self):
        jl = self.unpack
        for (
            character, rulebook, rule, thing, branch, turn, tick, handled_branch, handled_turn
        ) in self.sql('character_thing_rules_changes_dump'):
            yield (
                jl(character), jl(rulebook), rule, jl(thing),
                branch, turn, tick, handled_branch, handled_turn
            )

    def character_place_rules_handled_dump(self):
        for character, rulebook, rule, place, branch, turn, tick in self.sql('character_place_rules_handled_dump'):
            yield self.unpack(character), self.unpack(rulebook), rule, self.unpack(place), branch, turn, tick

    def character_place_rules_changes_dump(self):
        jl = self.unpack
        for (
            character, rulebook, rule, place, branch, turn, tick, handled_branch, handled_turn
        ) in self.sql('character_place_rules_changes_dump'):
            yield (
                jl(character), jl(rulebook), rule, jl(place),
                branch, turn, tick, handled_branch, handled_turn
            )

    def character_portal_rules_handled_dump(self):
        for character, rulebook, rule, orig, dest, branch, turn, tick in self.sql('character_portal_rules_handled_dump'):
            yield (
                self.unpack(character), self.unpack(rulebook), rule, self.unpack(orig), self.unpack(dest),
                branch, turn, tick
            )

    def character_portal_rules_changes_dump(self):
        jl = self.unpack
        for (
            character, rulebook, rule, orig, dest, branch, turn, tick, handled_branch, handled_turn
        ) in self.sql('character_portal_rules_changes_dump'):
            yield (
                jl(character), jl(rulebook), rule, jl(orig), jl(dest),
                branch, turn, tick, handled_branch, handled_turn
            )

    def node_rules_handled_dump(self):
        for character, node, rulebook, rule, branch, turn, tick in self.sql('node_rules_handled_dump'):
            yield self.unpack(character), self.unpack(node), self.unpack(rulebook), rule, branch, turn, tick

    def node_rules_changes_dump(self):
        jl = self.unpack
        for (
                character, node, rulebook, rule, branch, turn, tick, handled_branch, handled_turn
        ) in self.sql('node_rules_changes_dump'):
            yield (
                jl(character), jl(node), jl(rulebook), rule,
                branch, turn, tick, handled_branch, handled_turn
            )

    def portal_rules_handled_dump(self):
        for character, orig, dest, rulebook, rule, branch, turn, tick in self.sql('portal_rules_handled_dump'):
            yield (
                self.unpack(character), self.unpack(orig), self.unpack(dest),
                self.unpack(rulebook), rule, branch, turn, tick
            )

    def portal_rules_changes_dump(self):
        jl = self.unpack
        for (
            character, orig, dest, rulebook, rule, branch, turn, tick, handled_branch, handled_turn
        ) in self.sql('portal_rules_changes_dump'):
            yield (
                jl(character), jl(orig), jl(dest), jl(rulebook), rule,
                branch, turn, tick, handled_branch, handled_turn
            )

    def senses_dump(self):
        for character, sense, branch, turn, tick, function in self.sql('senses_dump'):
            yield self.unpack(character), sense, branch, turn, tick, function

    def things_dump(self):
        for character, thing, branch, turn, tick, location in self.sql('things_dump'):
            yield (
                self.unpack(character), self.unpack(thing), branch, turn, tick,
                self.unpack(location)
            )

    def avatars_dump(self):
        for character_graph, avatar_graph, avatar_node, branch, turn, tick, is_av in self.sql('avatars_dump'):
            yield (
                self.unpack(character_graph), self.unpack(avatar_graph),
                self.unpack(avatar_node), branch, turn, tick, is_av
            )

    def universal_set(self, key, branch, turn, tick, val):
        key, val = map(self.pack, (key, val))
        self.sql('universals_insert', key, branch, turn, tick, val)

    def universal_del(self, key, branch, turn, tick):
        key = self.pack(key)
        self.sql('universals_insert', key, branch, turn, tick, None)

    def comparison(
            self, entity0, stat0, entity1,
            stat1=None, oper='eq', windows=[]
    ):
        stat1 = stat1 or stat0
        return comparisons[oper](
            leftside=entity0.status(stat0),
            rightside=entity1.status(stat1),
            windows=windows
        )

    def count_all_table(self, tbl):
        return self.sql('{}_count'.format(tbl)).fetchone()[0]

    def init_table(self, tbl):
        try:
            return self.sql('create_{}'.format(tbl))
        except OperationalError:
            pass

    def rules_dump(self):
        for (name,) in self.sql('rules_dump'):
            yield name

    def _set_rule_something(self, what, rule, branch, turn, tick, flist):
        flist = self.pack(flist)
        return self.sql('rule_{}_insert'.format(what), rule, branch, turn, tick, flist)

    set_rule_triggers = partialmethod(_set_rule_something, 'triggers')
    set_rule_prereqs = partialmethod(_set_rule_something, 'prereqs')
    set_rule_actions = partialmethod(_set_rule_something, 'actions')

    def set_rule(self, rule, branch, turn, tick, triggers=None, prereqs=None, actions=None):
        self.sql('rules_insert', rule)
        self.set_rule_triggers(rule, branch, turn, tick, triggers or [])
        self.set_rule_prereqs(rule, branch, turn, tick, prereqs or [])
        self.set_rule_actions(rule, branch, turn, tick, actions or [])

    def set_rulebook(self, name, branch, turn, tick, rules=None):
        name, rules = map(self.pack, (name, rules or []))
        self.sql('rulebooks_insert', name, branch, turn, tick, rules)

    def _set_rulebook_on_character(self, rbtyp, char, branch, turn, tick, rb):
        char, rb = map(self.pack, (char, rb))
        self.sql(rbtyp + '_rulebook_insert', char, branch, turn, tick, rb)

    set_character_rulebook = partialmethod(_set_rulebook_on_character, 'character')
    set_avatar_rulebook = partialmethod(_set_rulebook_on_character, 'avatar')
    set_character_thing_rulebook = partialmethod(_set_rulebook_on_character, 'character_thing')
    set_character_place_rulebook = partialmethod(_set_rulebook_on_character, 'character_place')
    set_character_portal_rulebook = partialmethod(_set_rulebook_on_character, 'character_portal')

    def rulebooks(self):
        for book in self.sql('rulebooks'):
            yield self.unpack(book)

    def exist_node(self, character, node, branch, turn, tick, extant):
        super().exist_node(character, node, branch, turn, tick, extant)

    def exist_edge(self, character, orig, dest, idx, branch, turn, tick, extant=None):
        if extant is None:
            branch, turn, tick, extant = idx, branch, turn, tick
            idx = 0
        super().exist_edge(character, orig, dest, idx, branch, turn, tick, extant)

    def set_node_rulebook(self, character, node, branch, turn, tick, rulebook):
        (character, node, rulebook) = map(
            self.pack, (character, node, rulebook)
        )
        return self.sql('node_rulebook_insert', character, node, branch, turn, tick, rulebook)

    def set_portal_rulebook(self, character, orig, dest, branch, turn, tick, rulebook):
        (character, orig, dest, rulebook) = map(
            self.pack, (character, orig, dest, rulebook)
        )
        return self.sql(
            'portal_rulebook_insert',
            character,
            orig,
            dest,
            branch,
            turn,
            tick,
            rulebook
        )

    def handled_character_rule(
            self, character, rulebook, rule, branch, turn, tick
    ):
        (character, rulebook) = map(
            self.pack, (character, rulebook)
        )
        return self.sql(
            'character_rules_handled_insert',
            character,
            rulebook,
            rule,
            branch,
            turn,
            tick,
        )

    def handled_avatar_rule(self, character,  rulebook, rule, graph, av, branch, turn, tick):
        character, graph, av, rulebook = map(
            self.pack, (character, graph, av, rulebook)
        )
        return self.sql(
            'avatar_rules_handled_insert',
            character,
            rulebook,
            rule,
            graph,
            av,
            branch,
            turn,
            tick
        )

    def handled_character_thing_rule(self, character, rulebook, rule, thing, branch, turn, tick):
        character, thing, rulebook = map(
            self.pack, (character, thing, rulebook)
        )
        return self.sql(
            'character_thing_rules_handled_insert',
            character,
            rulebook,
            rule,
            thing,
            branch,
            turn,
            tick
        )

    def handled_character_place_rule(self, character, rulebook, rule, place, branch, turn, tick):
        character, rulebook, place = map(
            self.pack, (character, rulebook, place)
        )
        return self.sql(
            'character_place_rules_handled_insert',
            character,
            rulebook,
            rule,
            place,
            branch,
            turn,
            tick
        )

    def handled_character_portal_rule(self, character, rulebook, rule, orig, dest, branch, turn, tick):
        character, rulebook, orig, dest = map(
            self.pack, (character, rulebook, orig, dest)
        )
        return self.sql(
            'character_portal_rules_handled_insert',
            character,
            rulebook,
            rule,
            orig,
            dest,
            branch,
            turn,
            tick
        )

    def handled_node_rule(
            self, character, node, rulebook, rule, branch, turn, tick
    ):
        (character, node, rulebook) = map(
            self.pack, (character, node, rulebook)
        )
        return self.sql(
            'node_rules_handled_insert',
            character,
            node,
            rulebook,
            rule,
            branch,
            turn,
            tick
        )

    def handled_portal_rule(
            self, character, orig, dest, rulebook, rule, branch, turn, tick
    ):
        (character, orig, dest, rulebook) = map(
            self.pack, (character, orig, dest, rulebook)
        )
        return self.sql(
            'portal_rules_handled_insert',
            character,
            orig,
            dest,
            rulebook,
            rule,
            branch,
            turn,
            tick
        )

    def get_rulebook_char(self, rulemap, character):
        character = self.pack(character)
        for (book,) in self.sql(
                'rulebook_get_{}'.format(rulemap), character
        ):
            return self.unpack(book)
        raise KeyError("No rulebook")

    def set_thing_loc(
            self, character, thing, branch, turn, tick, loc
    ):
        (character, thing) = map(
            self.pack,
            (character, thing)
        )
        loc = self.pack(loc)
        self.sql('del_things_after', character, thing, branch, turn, turn, tick)
        self.sql(
            'things_insert',
            character,
            thing,
            branch,
            turn,
            tick,
            loc
        )

    def avatar_set(self, character, graph, node, branch, turn, tick, isav):
        (character, graph, node) = map(
            self.pack, (character, graph, node)
        )
        self.sql(
            'del_avatars_after',
            character, graph, node, branch, turn, turn, tick
        )
        self.sql(
            'avatars_insert', character, graph, node, branch, turn, tick, isav
        )

    def rulebooks_rules(self):
        for (rulebook, rule) in self.sql('rulebooks_rules'):
            yield map(self.unpack, (rulebook, rule))

    def rulebook_get(self, rulebook, idx):
        return self.unpack(
            self.sql(
                'rulebook_get', self.pack(rulebook), idx
            ).fetchone()[0]
        )

    def branch_descendants(self, branch):
        for child in self.sql('branch_children', branch):
            yield child
            yield from self.branch_descendants(child)

    def turns_completed_dump(self):
        return self.sql('turns_completed_dump')

    def complete_turn(self, branch, turn):
        try:
            self.sql('turns_completed_insert', branch, turn)
        except IntegrityError:
            self.sql('turns_completed_update', turn, branch)

    def initdb(self):
        """Set up the database schema, both for allegedb and the special
        extensions for LiSE

        """
        super().initdb()
        for table in (
            'universals',
            'rules',
            'rulebooks',
            'senses',
            'things',
            'character_rulebook',
            'avatar_rulebook',
            'character_thing_rulebook',
            'character_place_rulebook',
            'character_portal_rulebook',
            'node_rulebook',
            'portal_rulebook',
            'avatars',
            'character_rules_handled',
            'avatar_rules_handled',
            'character_thing_rules_handled',
            'character_place_rules_handled',
            'character_portal_rules_handled',
            'node_rules_handled',
            'portal_rules_handled',
            'rule_triggers',
            'rule_prereqs',
            'rule_actions',
            'turns_completed'
        ):
            self.init_table(table)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import sys
if sys.version_info[0] < 3 or (
        sys.version_info[0] == 3 and
        sys.version_info[1] < 3
):
    raise RuntimeError("LiSE requires Python 3.3 or later")

from setuptools import setup


setup(
    name="LiSE",
    version="0.9.1",
    description="Rules engine for life simulation games",
    author="Zachary Spector",
    author_email="zacharyspector@gmail.com",
    license="AGPL3+",
    keywords="game simulation",
    url="https://github.com/LogicalDash/LiSE",
    packages=[
        "LiSE",
        "LiSE.server",
        "LiSE.examples"
    ],
    package_data={
        'LiSE': ['sqlite.json']
    },
    install_requires=[
        "allegedb==0.14.0",
        "astunparse>=1.5.0",
        "u-msgpack-python>=2.4.1"
    ],
)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Wrap a LiSE engine so you can access and control it using only
ordinary method calls.

"""
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL
from re import match
from collections import defaultdict
from functools import partial
from importlib import import_module
from .engine import Engine


def dict_delta(old, new):
    """Return a dictionary containing the items of ``new`` that are either
    absent from ``old`` or whose values are different; as well as the
    value ``None`` for those keys that are present in ``old``, but
    absent from ``new``.

    Useful for describing changes between two versions of a dict.

    """
    r = {}
    oldkeys = set(old.keys())
    newkeys = set(new.keys())
    r.update((k, new[k]) for k in newkeys.difference(oldkeys))
    r.update((k, None) for k in oldkeys.difference(newkeys))
    for k in oldkeys.intersection(newkeys):
        if old[k] != new[k]:
            r[k] = new[k]
    return r


def set_delta(old, new):
    old = set(old)
    new = set(new)
    r = {}
    r.update((item, False) for item in old.difference(new))
    r.update((item, True) for item in new.difference(old))
    return r


def timely(fun):
    def run_timely(self, *args, **kwargs):
        ret = fun(self, *args, **kwargs)
        self.branch, self.turn, self.tick = self._real.btt()
        return ret

    run_timely.timely = True
    return run_timely


class EngineHandle(object):
    """A wrapper for a :class:`LiSE.Engine` object that runs in the same
    process, but with an API built to be used in a command-processing
    loop that takes commands from another process.

    It's probably a bad idea to use this class unless you're
    developing your own API.

    """
    def __init__(self, args, kwargs={}, logq=None, logfile=None, loglevel=None):
        """Instantiate an engine with the positional arguments ``args`` and
        the keyword arguments ``kwargs``.

        ``logq`` is a :class:`Queue` into which I'll put tuples of
        ``(loglevel, message)``.

        """
        kwargs.setdefault('logfun', self.log)
        self._real = Engine(*args, **kwargs)
        self._logq = logq
        self._loglevel = loglevel
        self._muted_chars = set()
        self.branch = self._real.branch
        self.turn = self._real.turn
        self.tick = self._real.tick
        self._node_stat_cache = defaultdict(dict)
        self._portal_stat_cache = defaultdict(
            lambda: defaultdict(dict)
        )
        self._char_stat_cache = {}
        self._char_av_cache = defaultdict(lambda: defaultdict(set))
        self._char_rulebooks_cache = {}
        self._char_nodes_rulebooks_cache = defaultdict(dict)
        self._char_portals_rulebooks_cache = defaultdict(
            lambda: defaultdict(dict)
        )
        self._char_nodes_cache = {}
        self._char_portals_cache = {}
        self._node_successors_cache = defaultdict(dict)
        self._strings_cache = {}
        self._eternal_cache = {}
        self._universal_cache = {}
        self._rule_cache = {}
        self._rulebook_cache = defaultdict(list)
        self._stores_cache = defaultdict(dict)

    def log(self, level, message):
        if isinstance(level, str):
            level = {
                'debug': 10,
                'info': 20,
                'warning': 30,
                'error': 40,
                'critical': 50
            }[level]
        if self._logq and level >= self._loglevel:
            self._logq.put((level, message))

    def debug(self, message):
        self.log(DEBUG, message)

    def info(self, message):
        self.log(INFO, message)

    def warning(self, message):
        self.log(WARNING, message)

    def error(self, message):
        self.log(ERROR, message)

    def critical(self, message):
        self.log(CRITICAL, message)

    def unpack(self, s):
        return self._real.unpack(s)

    def pack(self, o):
        return self._real.pack(o)

    def time_locked(self):
        return hasattr(self._real, 'locktime')

    @timely
    def advance(self):
        self._real.advance()

    def get_char_deltas(self, chars, *, store=True):
        ret = {}
        if chars == 'all':
            it = iter(self._real.character.keys())
        else:
            it = iter(chars)
        for char in it:
            delt = self.character_delta(char, store=store)
            if delt:
                ret[char] = delt
        return ret

    def _upd_local_caches(self, delta=None):
        if delta is None:
            self._eternal_cache = dict(self._real.eternal)
            self._universal_cache = dict(self._real.universal)
            self._rulebook_cache = {
                rb: self.rulebook_copy(rb) for rb in self._real.rulebook
            }
            self._rule_cache = {
                r: self.rule_copy(r) for r in self._real.rule
            }
            self._strings_cache = self.strings_copy()
            # stores?
            for charn, char in self._real.character.items():
                self._char_stat_cache[charn] = self.character_stat_copy(charn)
                self._node_stat_cache[charn] = {
                    node: self.node_stat_copy(charn, node) for node in char.node
                }
                portst = self._portal_stat_cache[charn]
                for port in char.portals():
                    portst.setdefault(port.orig, {})[port.dest] = self.portal_stat_copy(charn, port.orig, port.dest)
                self._char_nodes_cache[charn] = self.character_nodes(char)
                self._char_portals_cache[charn] = self.character_portals(char)
                self._char_rulebooks_cache[charn] = self.character_rulebooks_copy(char)
            return

        def updd(d0, d1):
            for k, v in d1.items():
                if v is None:
                    if k in d0:
                        del d0[k]
                else:
                    d0[k] = v
        updd(self._eternal_cache, delta.pop('eternal', {}))
        updd(self._universal_cache, delta.pop('universal', {}))
        updd(self._rulebook_cache, delta.pop('rulebooks', {}))
        updd(self._strings_cache, delta.pop('strings', {}))
        for rule, d in delta.pop('rules', {}).items():
            updd(self._rulebook_cache.setdefault(rule, {}), d)
        for char, d in delta.items():
            nodeset = set(self._char_nodes_cache.setdefault(char, ()))
            for n, ex in d.pop('nodes', {}).items():
                if ex:
                    nodeset.add(n)
                else:
                    nodeset.remove(n)
            nodevd = self._node_stat_cache.setdefault(char, {})
            for node, val in d.pop('node_val', {}).items():
                nodenvd = nodevd.setdefault(node, {})
                for k, v in val.items():
                    if k != 'location' and v is None:
                        if k in nodenvd:
                            del nodenvd[k]
                    else:
                        nodenvd[k] = v
            edges = set(self._char_portals_cache.setdefault(char, ()))
            for orig, dests in d.pop('edges', {}).items():
                for dest, exists in dests.items():
                    if exists:
                        edges.add((orig, dest))
                    else:
                        edges.remove((orig, dest))
            edgevd = self._portal_stat_cache.setdefault(char, {})
            for orig, dests in d.pop('edge_val', {}).items():
                for dest, val in dests.items():
                    updd(edgevd.setdefault(orig, {}).setdefault(dest, {}), val)

    @timely
    def next_turn(self):
        self.debug('calling next_turn at {}, {}, {}'.format(*self._real.btt()))
        ret, delta = self._real.next_turn()
        self._after_ret = partial(self._upd_local_caches, delta)
        return ret, delta

    def get_slow_delta(self, chars='all', store=True):
        delta = {}
        if chars:
            delta = self.get_char_deltas(chars, store=store)
        etd = self.eternal_delta(store=store)
        if etd:
            delta['eternal'] = etd
        unid = self.universal_delta(store=store)
        if unid:
            delta['universal'] = unid
        rud = self.all_rules_delta(store=store)
        if rud:
            delta['rules'] = rud
        rbd = self.all_rulebooks_delta(store=store)
        if rbd:
            delta['rulebooks'] = rbd
        return delta

    @timely
    def time_travel(self, branch, turn, tick=None, chars='all'):
        branch_from, turn_from, tick_from = self._real.btt()
        slow_delta = branch != branch_from
        self._real.time = (branch, turn)
        if tick is None:
            self.tick = tick = self._real.tick
        else:
            self._real.tick = tick
        self.branch = branch
        self.turn = turn
        if slow_delta:
            delta = self.get_slow_delta(chars)
        else:
            delta = self._real.get_delta(branch, turn_from, tick_from, turn, tick)
            self._after_ret = partial(self._upd_local_caches, delta)
        return None, delta

    @timely
    def increment_branch(self, chars=[]):
        branch = self._real.branch
        m = match('(.*)([0-9]+)', branch)
        if m:
            stem, n = m.groups()
            branch = stem + str(int(n)+1)
        else:
            branch += '1'
        if branch in self._real._branches:
            if m:
                n = int(n)
            else:
                stem = branch[:-1]
                n = 1
            while stem + str(n) in self._real._branches:
                n += 1
            branch = stem + str(n)
        ret = {'branch': branch}
        self._real.branch = self.branch = branch
        if chars:
            ret.update(self.get_char_deltas(chars))
        return ret

    @timely
    def add_character(self, char, data, attr):
        character = self._real.new_character(char, **attr)
        placedata = data.get('place', data.get('node', {}))
        for place, stats in placedata.items():
            character.add_place(place, **stats)
        thingdata = data.get('thing',  {})
        for thing, stats in thingdata.items():
            character.add_thing(thing, **stats)
        portdata = data.get('edge', data.get('portal', data.get('adj',  {})))
        for orig, dests in portdata.items():
            for dest, stats in dests.items():
                character.add_portal(orig, dest, **stats)

    def commit(self):
        self._real.commit()

    def close(self):
        self._real.close()

    def get_branch(self):
        return self._real.branch

    def get_watched_branch(self):
        return self.branch

    def set_branch(self, branch):
        self._real.branch = branch
        self.branch = branch

    def get_tick(self):
        return self._real.tick

    def get_watched_tick(self):
        return self.tick

    def set_tick(self, tick):
        self._real.tick = tick
        self.tick = tick

    def get_time(self):
        return self._real.time

    def get_watched_btt(self):
        return (self.branch, self.turn, self.tick)

    def get_language(self):
        return self._real.string.language

    def set_language(self, lang):
        self._real.string.language = lang
        return self.strings_delta()

    def get_string_ids(self):
        return list(self._real.string)

    def get_string_lang_items(self, lang):
        ret = list(self._real.string.lang_items(lang))
        for lang, (k, v) in ret:
            self._strings_cache[lang][k] = v
        return ret

    def strings_copy(self, lang=None):
        if lang is None:
            lang = self._real.string.language
        return dict(
            self._real.string.lang_items(lang)
        )

    def strings_delta(self):
        old = self._strings_cache
        new = dict(self._real.string)
        ret = dict_delta(old, new)
        self._strings_cache = new
        return ret

    def get_string(self, k):
        return self._real.string[k]

    def have_string(self, k):
        return k in self._real.string

    def set_string(self, k, v):
        self._real.string[k] = v
        self._strings_cache[k] = v

    def del_string(self, k):
        del self._real.string[k]
        del self._strings_cache[k]

    def get_eternal(self, k):
        ret = self._eternal_cache[k] = self._real.eternal[k]
        return ret

    def set_eternal(self, k, v):
        self._real.eternal[k] = v
        self._eternal_cache[k] = v

    def del_eternal(self, k):
        del self._real.eternal[k]
        del self._eternal_cache[k]

    def have_eternal(self, k):
        return k in self._real.eternal

    def eternal_copy(self):
        return dict(self._real.eternal)

    def eternal_delta(self, *, store=True):
        old = self._eternal_cache
        new = self.eternal_copy()
        if store:
            self._eternal_cache = new
        return dict_delta(old, new)

    def get_universal(self, k):
        ret = self._universal_cache[k] = self._real.universal[k]
        return ret

    @timely
    def set_universal(self, k, v):
        self._real.universal[k] = v
        self._universal_cache[k] = v

    @timely
    def del_universal(self, k):
        del self._real.universal[k]
        del self._universal_cache[k]

    def universal_copy(self):
        return dict(self._real.universal)

    def universal_delta(self, *, store=True):
        old = self._universal_cache
        new = self.universal_copy()
        if store:
            self._universal_cache = new
        return dict_delta(old, new)

    @timely
    def init_character(self, char, statdict={}):
        if char in self._real.character:
            raise KeyError("Already have character {}".format(char))
        self._real.character[char] = {}
        self._real.character[char].stat.update(statdict)

    @timely
    def del_character(self, char):
        del self._real.character[char]
        for cache in (
                self._char_stat_cache,
                self._node_stat_cache,
                self._portal_stat_cache,
                self._char_av_cache,
                self._char_rulebooks_cache,
                self._char_nodes_rulebooks_cache,
                self._char_portals_rulebooks_cache,
                self._char_nodes_cache,
                self._char_portals_cache
        ):
            if char in cache:
                del cache[char]

    def character_stat_copy(self, char):
        return {
            k: v.unwrap() if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap') else v
            for (k, v) in self._real.character[char].stat.items()
        }

    @staticmethod
    def _character_something_delta(char, cache, copier, *args, store=True):
        old = cache.get(char, {})
        new = copier(char, *args)
        if store:
            cache[char] = new
        return dict_delta(old, new)

    def character_stat_delta(self, char, *, store=True):
        return self._character_something_delta(
            char, self._char_stat_cache, self.character_stat_copy, store=store
        )

    def character_avatars_copy(self, char):
        return {
            graph: frozenset(nodes.keys()) for (graph, nodes) in
            self._real.character[char].avatar.items()
        }

    def character_avatars_delta(self, char, *, store=True):
        old = self._char_av_cache.get(char, {})
        new = self.character_avatars_copy(char)
        ret = {}
        oldkeys = set(old.keys())
        newkeys = set(new.keys())
        for graph in oldkeys.difference(newkeys):
            ret[graph] = {node: False for node in old[graph]}
        for graph in newkeys.difference(oldkeys):
            ret[graph] = {node: True for node in new[graph]}
        for graph in oldkeys.intersection(newkeys):
            graph_nodes = {}
            for node in old[graph].difference(new[graph]):
                graph_nodes[node] = False
            for node in new[graph].difference(old[graph]):
                graph_nodes[node] = True
            if graph_nodes:
                ret[graph] = graph_nodes
        if store:
            self._char_av_cache[char] = new
        return ret

    def character_rulebooks_copy(self, char):
        chara = self._real.character[char]
        return {
            'character': chara.rulebook.name,
            'avatar': chara.avatar.rulebook.name,
            'thing': chara.thing.rulebook.name,
            'place': chara.place.rulebook.name,
            'portal': chara.portal.rulebook.name
        }

    def character_rulebooks_delta(self, char, *, store=True):
        return self._character_something_delta(
            char, self._char_rulebooks_cache, self.character_rulebooks_copy, store=store
        )

    def character_nodes_rulebooks_copy(self, char, nodes='all'):
        chara = self._real.character[char]
        if nodes == 'all':
            nodeiter = iter(chara.node.values())
        else:
            nodeiter = (chara.node[k] for k in nodes)
        return {node.name: node.rulebook.name for node in nodeiter}

    def character_nodes_rulebooks_delta(self, char, nodes='all', *, store=True):
        return self._character_something_delta(
            char, self._char_nodes_rulebooks_cache,
            self.character_nodes_rulebooks_copy, nodes, store=store
        )

    def character_portals_rulebooks_copy(self, char, portals='all'):
        chara = self._real.character[char]
        result = defaultdict(dict)
        if portals == 'all':
            portiter = chara.portals()
        else:
            portiter = (chara.portal[orig][dest] for (orig, dest) in portals)
        for portal in portiter:
            result[portal['origin']][portal['destination']] \
                = portal.rulebook.name
        return result

    def character_portals_rulebooks_delta(self, char, portals='all', *, store=True):
        try:
            old = self._char_portals_rulebooks_cache.get(
                char, defaultdict(dict)
            )
            new = self.character_portals_rulebooks_copy(char, portals)
            if store:
                self._char_portals_rulebooks_cache[char] = new
            result = {}
            for origin in old:
                if origin in new:
                    result[origin] = dict_delta(old[origin], new[origin])
                else:
                    result[origin] = None
            for origin in new:
                if origin not in result:
                    result[origin] = new[origin]
            return result
        except KeyError:
            return None

    def character_copy(self, char):
        """Return a dictionary describing character ``char``."""
        ret = self.character_stat_copy(char)
        chara = self._real.character[char]
        nv = self.character_nodes_stat_copy(char)
        if nv:
            ret['node_val'] = nv
        ev = self.character_portals_stat_copy(char)
        if ev:
            ret['edge_val'] = ev
        avs = self.character_avatars_copy(char)
        if avs:
            ret['avatars'] = avs
        rbs = self.character_rulebooks_copy(char)
        if rbs:
            ret['rulebooks'] = rbs
        nrbs = self.character_nodes_rulebooks_copy(char)
        if nrbs:
            for node, rb in nrbs.items():
                assert node in chara.node
                if 'node_val' not in ret:
                    ret['node_val'] = {}
                nv = ret['node_val']
                if node not in nv:
                    nv[node] = {}
                nv[node]['rulebook'] = rb
        porbs = self.character_portals_rulebooks_copy(char)
        if porbs:
            for orig, dests in porbs.items():
                ports = chara.portal[orig]
                for dest, rb in dests.items():
                    assert dest in ports
                    if 'edge_val' not in ret:
                        ret['edge_val'] = {}
                    ev = ret['edge_val']
                    if orig not in ev:
                        ev[orig] = {}
                    ov = ev[orig]
                    if dest not in ov:
                        ov[dest] = {}
                    ov[dest]['rulebook'] = rb
        return ret

    def character_delta(self, char, *, store=True):
        """Return a dictionary of changes to ``char`` since previous call."""
        ret = self.character_stat_delta(char, store=store)
        nodes = self.character_nodes_delta(char, store=store)
        chara = self._real.character[char]
        if nodes:
            ret['nodes'] = nodes
        edges = self.character_portals_delta(char, store=store)
        if edges:
            ret['edges'] = edges
        avs = self.character_avatars_delta(char, store=store)
        if avs:
            ret['avatars'] = avs
        rbs = self.character_rulebooks_delta(char, store=store)
        if rbs:
            ret['rulebooks'] = rbs
        nrbs = self.character_nodes_rulebooks_delta(char, store=store)
        if nrbs:
            for node, rb in nrbs.items():
                if node not in chara.node:
                    continue
                ret.setdefault('node_val', {}).setdefault(node, {})['rulebook'] = rb
        porbs = self.character_portals_rulebooks_delta(char, store=store)
        if porbs:
            for orig, dests in porbs.items():
                if orig not in chara.portal:
                    continue
                portals = chara.portal[orig]
                for dest, rb in dests.items():
                    if dest not in portals:
                        continue
                    ret.setdefault('edge_val', {}).setdefault(orig, {}).setdefault(dest, {})['rulebook'] = rb
        nv = self.character_nodes_stat_delta(char, store=store)
        if nv:
            ret['node_val'] = nv
        ev = self.character_portals_stat_delta(char, store=store)
        if ev:
            ret['edge_val'] = ev
        return ret

    @timely
    def set_character_stat(self, char, k, v):
        self._real.character[char].stat[k] = v
        self._char_stat_cache.setdefault(char, {})[k] = v

    @timely
    def del_character_stat(self, char, k):
        del self._real.character[char].stat[k]
        del self._char_stat_cache[char][k]

    @timely
    def update_character_stats(self, char, patch):
        self._real.character[char].stat.update(patch)
        self._char_stat_cache.setdefault(char, {}).update(patch)

    @timely
    def update_character(self, char, patch):
        self.update_character_stats(char, patch['character'])
        self.update_nodes(char, patch['node'])
        self.update_portals(char, patch['portal'])

    def characters(self):
        return list(self._real.character.keys())

    @timely
    def set_character(self, char, v):
        self._real.character[char] = v

    @timely
    def set_node_stat(self, char, node, k, v):
        self._real.character[char].node[node][k] = v
        self._node_stat_cache.setdefault(char, {})[node][k] = v

    @timely
    def del_node_stat(self, char, node, k):
        del self._real.character[char].node[node][k]
        del self._node_stat_cache[char][node][k]

    def node_stat_copy(self, node_or_char, node=None):
        """Return a node's stats, prepared for pickling, in a dictionary."""
        if node is None:
            node = node_or_char
        else:
            node = self._real.character[node_or_char].node[node]
        return {
            k: v.unwrap() if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap') else v
            for (k, v) in node.items() if k not in {
                'character',
                'name',
                'arrival_time',
                'next_arrival_time'
            }
        }

    def node_stat_delta(self, char, node, *, store=True):
        """Return a dictionary describing changes to a node's stats since the
        last time you looked at it.

        Deleted keys have the value ``None``. If the node's been deleted, this
        returns ``None``.

        """
        try:
            old = self._node_stat_cache[char].get(node, {})
            new = self.node_stat_copy(self._real.character[char].node[node])
            if store:
                self._node_stat_cache[char][node] = new
            r = dict_delta(old, new)
            return r
        except KeyError:
            return None

    def character_nodes_stat_delta(self, char, *, store=True):
        """Return a dictionary of ``node_stat_delta`` output for each node in a
        character.

        """
        r = {}
        nodes = set(self._real.character[char].node.keys())
        for node in nodes:
            delta = self.node_stat_delta(char, node, store=store)
            if delta:
                r[node] = delta
        nsc = self._node_stat_cache[char]
        for node in list(nsc.keys()):
            if node not in nodes:
                del nsc[node]
        return r

    def character_nodes_stat_copy(self, char):
        return {node: self.node_stat_copy(char, node) for node in self._real.character[char].node}

    @timely
    def update_node(self, char, node, patch):
        """Change a node's stats according to a dictionary.

        The ``patch`` dictionary should hold the new values of stats,
        keyed by the stats' names; a value of ``None`` deletes the
        stat.

        """
        character = self._real.character[char]
        if patch is None:
            del character.node[node]
        elif node not in character.node:
            character.node[node] = patch
            return
        else:
            character.node[node].update(patch)

    @timely
    def update_nodes(self, char, patch, backdate=False):
        """Change the stats of nodes in a character according to a
        dictionary.

        """
        if backdate:
            parbranch, parrev = self._real._parentbranch_rev.get(
                self._real.branch, ('trunk', 0)
            )
            tick_now = self._real.tick
            self._real.tick = parrev
        for i, (n, npatch) in enumerate(patch.items(), 1):
            self.update_node(char, n, npatch)
        if backdate:
            self._real.tick = tick_now

    @timely
    def del_node(self, char, node):
        """Remove a node from a character."""
        del self._real.character[char].node[node]
        for cache in (
                self._char_nodes_rulebooks_cache,
                self._node_stat_cache,
                self._node_successors_cache
        ):
            try:
                del cache[char][node]
            except KeyError:
                pass
        if char in self._char_nodes_cache and node in self._char_nodes_cache[char]:
            self._char_nodes_cache[char] = self._char_nodes_cache[char] - frozenset([node])
        if char in self._portal_stat_cache:
            portal_stat_cache_char = self._portal_stat_cache[char]
            if node in portal_stat_cache_char:
                del portal_stat_cache_char[node]
            for charo in portal_stat_cache_char.values():
                if node in charo:
                    del charo[node]
        if char in self._char_portals_rulebooks_cache:
            portal_rulebook_cache_char = self._char_portals_rulebooks_cache[char]
            if node in portal_rulebook_cache_char:
                del portal_rulebook_cache_char[node]
            for porto in portal_rulebook_cache_char.values():
                if node in porto:
                    del porto[node]

    def character_nodes(self, char):
        return frozenset(self._real.character[char].node)

    def character_nodes_delta(self, char, *, store=True):
        try:
            old = self._char_nodes_cache.get(char, [])
            new = self.character_nodes(char)
            if store:
                self._char_nodes_cache[char] = new
            return set_delta(old, new)
        except KeyError:
            return None

    def node_predecessors(self, char, node):
        return list(self._real.character[char].pred[node].keys())

    def character_set_node_predecessors(self, char, node, preds):
        self._real.character[char].pred[node] = preds

    def character_del_node_predecessors(self, char, node):
        del self._real.character[char].pred[node]

    def node_successors(self, char, node):
        return list(self._real.character[char].portal[node].keys())

    def node_successors_delta(self, char, node):
        try:
            old = self._node_successors_cache[char].get(node, [])
            new = self.node_successors(char, node)
            self._node_successors_cache[char][node] = new
            return set_delta(old, new)
        except KeyError:
            return None

    def character_set_node_successors(self, char, node, val):
        self._real.character[char].adj[node] = val

    def character_del_node_successors(self, char, node):
        del self._real.character[char].adj[node]

    def nodes_connected(self, char, orig, dest):
        return dest in self._real.character[char].portal[orig]

    @timely
    def init_thing(self, char, thing, statdict={}):
        if thing in self._real.character[char].thing:
            raise KeyError(
                'Already have thing in character {}: {}'.format(
                    char, thing
                )
            )
        return self.set_thing(char, thing, statdict)

    @timely
    def set_thing(self, char, thing, statdict):
        self._real.character[char].thing[thing] = statdict
        self._node_stat_cache.setdefault(char, {})[thing] = statdict
        if char in self._char_nodes_cache:
            cache = self._char_nodes_cache[char]
        else:
            cache = frozenset()
        self._char_nodes_cache[char] = cache.union((thing,))

    @timely
    def add_thing(self, char, thing, loc, statdict):
        self._real.character[char].add_thing(
            thing, loc, **statdict
        )
        self._node_stat_cache.setdefault(char, {})[thing] = statdict
        if char in self._char_nodes_cache:
            cache = self._char_nodes_cache[char]
        else:
            cache = frozenset()
        self._char_nodes_cache[char] = cache.union((thing,))

    @timely
    def place2thing(self, char, node, loc):
        self._real.character[char].place2thing(node, loc)

    @timely
    def thing2place(self, char, node):
        self._real.character[char].thing2place(node)

    @timely
    def add_things_from(self, char, seq):
        self._real.character[char].add_things_from(seq)

    def get_thing_location(self, char, thing):
        try:
            return self._real.character[char].thing[thing]['location']
        except KeyError:
            return None

    @timely
    def set_thing_location(self, char, thing, loc):
        self._real.character[char].thing[thing]['location'] = loc
        self._node_stat_cache.setdefault(char, {}).setdefault(thing, {})['location'] = loc

    @timely
    def thing_follow_path(self, char, thing, path, weight):
        return self._real.character[char].thing[thing].follow_path(path, weight)

    @timely
    def thing_go_to_place(self, char, thing, place, weight):
        return self._real.character[char].thing[thing].go_to_place(place, weight)

    @timely
    def thing_travel_to(self, char, thing, dest, weight=None, graph=None):
        """Make something find a path to ``dest`` and follow it.

        Optional argument ``weight`` is the portal stat to use to schedule movement times.

        Optional argument ``graph`` is an alternative graph to use for pathfinding.
        Should resemble a networkx DiGraph.

        """
        return self._real.character[char].thing[thing].travel_to(dest, weight, graph)

    @timely
    def init_place(self, char, place, statdict={}):
        if place in self._real.character[char].place:
            raise KeyError(
                'Already have place in character {}: {}'.format(
                    char, place
                )
            )
        return self.set_place(char, place, statdict)

    @timely
    def set_place(self, char, place, statdict):
        self._real.character[char].place[place] = statdict
        self._node_stat_cache.setdefault(char, {})[place] = statdict

    @timely
    def add_places_from(self, char, seq):
        self._real.character[char].add_places_from(seq)

    @timely
    def init_portal(self, char, orig, dest, statdict={}):
        if (
                orig in self._real.character[char].portal and
                dest in self._real.character[char].portal[orig]
        ):
            raise KeyError(
                'Already have portal in character {}: {}->{}'.format(
                    char, orig, dest
                )
            )
        return self.set_portal(char, orig, dest, statdict)

    @timely
    def set_portal(self, char, orig, dest, statdict):
        self._real.character[char].portal[orig][dest] = statdict
        self._portal_stat_cache.setdefault(char, {})[orig][dest] = statdict

    def character_portals(self, char):
        r = set()
        portal = self._real.character[char].portal
        for o in portal:
            for d in portal[o]:
                r.add((o, d))
        return r

    def character_portals_delta(self, char, *, store=True):
        old = self._char_portals_cache.get(char, {})
        new = self.character_portals(char)
        if store:
            self._char_portals_cache[char] = new
        ret = {}
        for orig, dest in old:
            if (orig, dest) not in new:
                ret.setdefault(orig, {})[dest] = False
                if store:
                    try:
                        del self._portal_stat_cache[char][orig][dest]
                    except KeyError:
                        pass
        for orig, dest in new:
            if (orig, dest) not in old:
                ret.setdefault(orig, {})[dest] = True
        return ret

    @timely
    def add_portal(self, char, orig, dest, symmetrical, statdict):
        self._real.character[char].add_portal(
            orig, dest, symmetrical, **statdict
        )

    @timely
    def add_portals_from(self, char, seq, symmetrical):
        self._real.character[char].add_portals_from(seq, symmetrical)

    def del_portal(self, char, orig, dest):
        del self._real.character[char].portal[orig][dest]
        try:
            del self._portal_stat_cache[char][orig][dest]
        except KeyError:
            pass

    def set_portal_stat(self, char, orig, dest, k, v):
        self._real.character[char].portal[orig][dest][k] = v
        self._portal_stat_cache.setdefault(char, {}).setdefault(orig, {}).setdefault(dest, {})[k] = v

    def del_portal_stat(self, char, orig, dest, k):
        del self._real.character[char][orig][dest][k]
        del self._portal_stat_cache[char][orig][dest][k]

    def portal_stat_copy(self, char, orig, dest):
        return {
            k: v.unwrap() if hasattr(v, 'unwrap') and not hasattr(v, 'no_unwrap') else v
            for (k, v) in self._real.character[char].portal[orig][dest].items()
        }

    def portal_stat_delta(self, char, orig, dest, *, store=True):
        try:
            old = self._portal_stat_cache[char][orig].get(dest, {})
            new = self.portal_stat_copy(char, orig, dest)
            if store:
                self._portal_stat_cache[char][orig][dest] = new
            return dict_delta(old, new)
        except KeyError:
            return None

    def character_portals_stat_copy(self, char):
        r = {}
        chara = self._real.character[char]
        for orig, dests in chara.portal.items():
            for dest in dests:
                if orig not in r:
                    r[orig] = {}
                r[orig][dest] = self.portal_stat_copy(char, orig, dest)
        return r

    def character_portals_stat_delta(self, char, *, store=True):
        r = {}
        for orig in self._real.character[char].portal:
            for dest in self._real.character[char].portal[orig]:
                delta = self.portal_stat_delta(char, orig, dest, store=store)
                if delta:
                    if orig not in r:
                        r[orig] = {}
                    r[orig][dest] = delta
        return r

    @timely
    def update_portal(self, char, orig, dest, patch):
        character = self._real.character[char]
        if patch is None:
            del character.portal[orig][dest]
        elif orig not in character.portal \
             or dest not in character.portal[orig]:
            character.portal[orig][dest] = patch
        else:
            character.portal[orig][dest].update(patch)

    @timely
    def update_portals(self, char, patch):
        branch = self.branch
        for ((orig, dest), ppatch) in patch.items():
            branch = self.update_portal(char, orig, dest, ppatch)
        return branch

    @timely
    def add_avatar(self, char, graph, node):
        self._real.character[char].add_avatar(graph, node)
        self._char_av_cache.setdefault(char, {})[graph].add(node)

    @timely
    def del_avatar(self, char, graph, node):
        self._real.character[char].del_avatar(graph, node)
        self._char_av_cache.setdefault(char, {})[graph].remove(node)

    @timely
    def new_empty_rule(self, rule):
        self._real.rule.new_empty(rule)

    @timely
    def new_empty_rulebook(self, rulebook):
        self._real.rulebook[rulebook]

    def rulebook_copy(self, rulebook):
        return list(self._real.rulebook[rulebook]._get_cache(*self._real.btt()))

    def rulebook_delta(self, rulebook, *, store=True):
        old = self._rulebook_cache[rulebook]
        new = self.rulebook_copy(rulebook)
        if store:
            self._rulebook_cache[rulebook] = new
        if old == new:
            return
        return new

    def all_rulebooks_delta(self, *, store=True):
        ret = {}
        for rulebook in self._real.rulebook.keys():
            delta = self.rulebook_delta(rulebook, store=store)
            if delta:
                ret[rulebook] = delta
        return ret

    @timely
    def set_rulebook_rule(self, rulebook, i, rule):
        self._real.rulebook[rulebook][i] = rule


    @timely
    def ins_rulebook_rule(self, rulebook, i, rule):
        self._real.rulebook[rulebook].insert(i, rule)

    @timely
    def del_rulebook_rule(self, rulebook, i):
        del self._real.rulebook[rulebook][i]
        del self._rulebook_cache[rulebook][i]

    @timely
    def set_rule_triggers(self, rule, triggers):
        self._real.rule[rule].triggers = triggers

    @timely
    def set_rule_prereqs(self, rule, prereqs):
        self._real.rule[rule].prereqs = prereqs

    @timely
    def set_rule_actions(self, rule, actions):
        self._real.rule[rule].actions = actions

    @timely
    def set_character_rulebook(self, char, rulebook):
        self._real.character[char].rulebook = rulebook

    @timely
    def set_avatar_rulebook(self, char, rulebook):
        self._real.character[char].avatar.rulebook = rulebook

    @timely
    def set_character_thing_rulebook(self, char, rulebook):
        self._real.character[char].thing.rulebook = rulebook

    @timely
    def set_character_place_rulebook(self, char, rulebook):
        self._real.character[char].place.rulebook = rulebook

    @timely
    def set_character_node_rulebook(self, char, rulebook):
        self._real.character[char].node.rulebook = rulebook

    @timely
    def set_character_portal_rulebook(self, char, rulebook):
        self._real.character[char].portal.rulebook = rulebook

    @timely
    def set_node_rulebook(self, char, node, rulebook):
        self._real.character[char].node[node].rulebook = rulebook

    @timely
    def set_portal_rulebook(self, char, orig, dest, rulebook):
        self._real.character[char].portal[orig][dest].rulebook = rulebook

    def rule_copy(self, rule):
        branch, turn, tick = self.branch, self.turn, self.tick
        return {
            'triggers': list(self._real._triggers_cache.retrieve(rule, branch, turn, tick)),
            'prereqs': list(self._real._prereqs_cache.retrieve(rule, branch, turn, tick)),
            'actions': list(self._real._actions_cache.retrieve(rule, branch, turn, tick))
        }

    def rule_delta(self, rule, *, store=True):
        old = self._rule_cache.get(rule, {'triggers': [], 'prereqs': [], 'actions': []})
        new = self.rule_copy(rule)
        if store:
            self._rule_cache[rule] = new
        ret = {}
        if new['triggers'] != old['triggers']:
            ret['triggers'] = new['triggers']
        if new['prereqs'] != old['prereqs']:
            ret['prereqs'] = new['prereqs']
        if new['actions'] != old['actions']:
            ret['actions'] = new['actions']
        return ret

    def all_rules_delta(self, *, store=True):
        ret = {}
        for rule in self._real.rule.keys():
            try:
                delta = self.rule_delta(rule, store=store)
                if delta:
                    ret[rule] = delta
            except KeyError:
                pass
        return ret

    def source_copy(self, store):
        return dict(getattr(self._real, store).iterplain())

    def source_delta(self, store):
        old = self._stores_cache.get(store, {})
        new = self._stores_cache[store] = self.source_copy(store)
        return dict_delta(old, new)

    def get_source(self, store, name):
        return getattr(self._real, store).get_source(name)

    def store_source(self, store, v, name=None):
        getattr(self._real, store).store_source(v, name)
        self._stores_cache.setdefault(store, {})[name or v.__name__] = v

    def del_source(self, store, k):
        delattr(getattr(self._real, store), k)
        try:
            del self._stores_cache[store][k]
        except KeyError:
            pass

    @timely
    def call_stored_function(self, store, func, args, kwargs):
        if store == 'method':
            args = (self._real,) + tuple(args)
        store = getattr(self._real, store)
        if store not in self._real.stores:
            raise ValueError("{} is not a function store".format(store))
        callme = getattr(store, func)
        try:
            return callme(*args, **kwargs)
        except Exception as ex:
            raise

    @timely
    def call_randomizer(self, method, *args, **kwargs):
        return getattr(self._real._rando, method)(*args, **kwargs)

    @timely
    def install_module(self, module):
        import_module(module).install(self._real)

    @timely
    def do_game_start(self):
        self._real.game_start()

    def is_parent_of(self, parent, child):
        return self._real.is_parent_of(parent, child)


# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Exception classes for use in LiSE."""
from allegedb.query import IntegrityError, TimeError

try:
    from sqlalchemy.exc import OperationalError as alchemyOpError
    from sqlite3 import OperationalError as liteOpError
    OperationalError = (alchemyOpError, liteOpError)
except ImportError:
    from sqlite3 import OperationalError


class NonUniqueError(Exception):
    """For when you tried to look up the only one of something but there wasn't just one"""


class AmbiguousAvatarError(NonUniqueError, KeyError):
    """An AvatarMapping can't decide what you want."""


class AmbiguousUserError(NonUniqueError, AttributeError):
    """A user descriptor can't decide what you want."""


class UserFunctionError(SyntaxError):
    """Error condition for when I try to load a user-defined function and
    something goes wrong.

    """
    pass


class WorldIntegrityError(ValueError):
    """Error condition for when something breaks the world model, even if
    it might be allowed by the database schema.

    """


class CacheError(ValueError):
    """Error condition for something going wrong with a cache"""
    pass


class TravelException(Exception):
    """Exception for problems with pathfinding. Not necessarily an error
    because sometimes somebody SHOULD get confused finding a path.

    """
    def __init__(
            self,
            message,
            path=None,
            followed=None,
            traveller=None,
            branch=None,
            turn=None,
            lastplace=None
    ):
        """Store the message as usual, and also the optional arguments:

        ``path``: a list of Place names to show such a path as you found

        ``followed``: the portion of the path actually followed

        ``traveller``: the Thing doing the travelling

        ``branch``: branch during travel

        ``tick``: tick at time of error (might not be the tick at the
        time this exception is raised)

        ``lastplace``: where the traveller was, when the error happened

        """
        self.path = path
        self.followed = followed
        self.traveller = traveller
        self.branch = branch
        self.turn = turn
        self.lastplace = lastplace
        super().__init__(message)


class PlanError(AttributeError):
    """Tried to use an attribute that shouldn't be used while planning"""


class RulesEngineError(Exception):
    """For problems to do with the rules engine

    Rules themselves should never raise this. Only the engine should.

    """


class RuleError(RulesEngineError):
    """For problems to do with rules

    Rather than the operation of the rules engine as a whole.

    Don't use this in your trigger, prereq, or action functions.
    It's only for Rule objects as such.

    """


class RedundantRuleError(RuleError):
    """Error condition for when you try to run a rule on a (branch,
    turn) it's already been executed.

    """